/* GStreamer
 * Copyright (C) 2008 Wim Taymans <wim.taymans at gmail.com>

* Copyright (c) 2012, 2013 Samsung Electronics Co., Ltd.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *

* * Modifications by Samsung Electronics Co., Ltd.

* 1. Applied Miracast WFD Server function
*/

#include <gst/gst.h>
#include <gst/rtsp/gstrtspconnection.h>


#ifndef __GST_RTSP_CLIENT_H__
#define __GST_RTSP_CLIENT_H__

G_BEGIN_DECLS

typedef struct _GstRTSPClient GstRTSPClient;
typedef struct _GstRTSPClientClass GstRTSPClientClass;
typedef struct _GstRTSPClientState GstRTSPClientState;
typedef struct _GstRTSPClientSrcBin GstRTSPClientSrcBin;

#include "wfdconfigmessage.h"
#include "rtsp-server.h"
#include "rtsp-media.h"
#include "rtsp-media-mapping.h"
#include "rtsp-session-pool.h"
#include "rtsp-auth.h"

#define GST_TYPE_RTSP_CLIENT              (gst_rtsp_client_get_type ())
#define GST_IS_RTSP_CLIENT(obj)           (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GST_TYPE_RTSP_CLIENT))
#define GST_IS_RTSP_CLIENT_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), GST_TYPE_RTSP_CLIENT))
#define GST_RTSP_CLIENT_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), GST_TYPE_RTSP_CLIENT, GstRTSPClientClass))
#define GST_RTSP_CLIENT(obj)              (G_TYPE_CHECK_INSTANCE_CAST ((obj), GST_TYPE_RTSP_CLIENT, GstRTSPClient))
#define GST_RTSP_CLIENT_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), GST_TYPE_RTSP_CLIENT, GstRTSPClientClass))
#define GST_RTSP_CLIENT_CAST(obj)         ((GstRTSPClient*)(obj))
#define GST_RTSP_CLIENT_CLASS_CAST(klass) ((GstRTSPClientClass*)(klass))

typedef enum {
  CLIENT_STATE_UNKNOWN = -1,
  CLIENT_STATE_CONNECTING = 0,
  CLIENT_STATE_CONNECTED = 1,
}ClientState;

typedef enum {
  WFD_MESSAGE_UNKNOWN,
  WFD_MESSAGE_1,
  WFD_MESSAGE_2,
  WFD_MESSAGE_3,
  WFD_MESSAGE_4,
  WFD_MESSAGE_5,
  WFD_MESSAGE_6,
  WFD_MESSAGE_7,
  WFD_MESSAGE_8,
  WFD_MESSAGE_9,
  WFD_MESSAGE_10,
  WFD_MESSAGE_11,
  WFD_MESSAGE_12,
  WFD_MESSAGE_13,
  WFD_MESSAGE_14,
  WFD_MESSAGE_15,
  WFD_MESSAGE_16,
}WFDMessageType;

typedef enum __wfd_ini_videosink_element
{
  WFD_INI_VSINK_V4l2SINK = 0,
  WFD_INI_VSINK_XIMAGESINK,
  WFD_INI_VSINK_XVIMAGESINK,
  WFD_INI_VSINK_FAKESINK,
  WFD_INI_VSINK_EVASIMAGESINK,
  WFD_INI_VSINK_GLIMAGESINK,
  WFD_INI_VSINK_NUM
}WFD_INI_VSINK_ELEMENT;

typedef enum __wfd_ini_videosrc_element
{
  WFD_INI_VSRC_XVIMAGESRC,
  WFD_INI_VSRC_FILESRC,
  WFD_INI_VSRC_CAMERASRC,
  WFD_INI_VSRC_VIDEOTESTSRC,
  WFD_INI_VSRC_NUM
}WFD_INI_VSRC_ELEMENT;

typedef enum __wfd_ini_session_mode
{
  WFD_INI_AUDIO_VIDEO_MUXED,
  WFD_INI_VIDEO_ONLY,
  WFD_INI_AUDIO_ONLY,
  WFD_INI_AUDIO_VIDEO_SAPERATE
}WFD_INI_SESSION_MODE;

/**
 * GstRTSPClientState:
 * @request: the complete request
 * @uri: the complete url parsed from @request
 * @method: the parsed method of @uri
 * @session: the session, can be NULL
 * @sessmedia: the session media for the url can be NULL
 * @factory: the media factory for the url, can be NULL.
 * @media: the session media for the url can be NULL
 * @response: the response
 *
 * Information passed around containing the client state of a request.
 */
struct _GstRTSPClientState{
  GstRTSPMessage      *request;
  GstRTSPUrl          *uri;
  GstRTSPMethod        method;
  GstRTSPSession      *session;
  GstRTSPSessionMedia *sessmedia;
  GstRTSPMediaFactory *factory;
  GstRTSPMedia        *media;
  GstRTSPMessage      *response;
};

struct _GstRTSPClientSrcBin{
  GstBin *srcbin;
  GstElement *vqueue;
  GstElement *aqueue;
  GstElement *tee_0;
  GstElement *venc;
  GstPad *mux_vsinkpad;
  GstPad *mux_asinkpad;
  GstElement *vparse;

};

/**
 * GstRTSPClient:
 *
 * @connection: the connection object handling the client request.
 * @watch: watch for the connection
 * @watchid: id of the watch
 * @ip: ip address used by the client to connect to us
 * @session_pool: handle to the session pool used by the client.
 * @media_mapping: handle to the media mapping used by the client.
 * @uri: cached uri
 * @media: cached media
 * @streams: a list of streams using @connection.
 * @sessions: a list of sessions managed by @connection.
 *
 * The client structure.
 */
struct _GstRTSPClient {
  GObject       parent;

  GstRTSPConnection *connection;
  GstRTSPWatch      *watch;
  guint            watchid;
  gchar           *server_ip;
  gboolean	is_ipv6;
  gchar		*wfdsink_ip;

  GstRTSPServer        *server;
  GstRTSPSessionPool   *session_pool;
  gchar * sessionid;
  GstRTSPMediaMapping  *media_mapping;
  GstRTSPAuth          *auth;

  GstRTSPUrl     *uri;
  GstRTSPMedia   *media;

  GList *streams;
  GList *sessions;

  /* supported methods */
  gint               methods;

  GstRTSPClientSrcBin *srcbin;

  int videosrc_type;
  gint session_mode;
  gchar *infile;

  GMutex *state_lock;
  GCond *state_wait;
  ClientState client_state;

  WFDAudioFormats caCodec;
  WFDAudioFreq cFreq;
  WFDAudioChannels cChanels;
  guint cBitwidth;
  guint caLatency;
  WFDVideoCodecs cvCodec;
  WFDVideoNativeResolution cNative;
  guint64 cNativeResolution;
  WFDVideoCEAResolution cCEAResolution;
  WFDVideoVESAResolution cVESAResolution;
  WFDVideoHHResolution cHHResolution;
  WFDVideoH264Profile cProfile;
  WFDVideoH264Level cLevel;
  guint32 cMaxHeight;
  guint32 cMaxWidth;
  guint32 cmin_slice_size;
  guint32 cslice_enc_params;
  guint cframe_rate_control;
  guint bitrate;
  guint MTUsize;
  guint cvLatency;
  WFDRTSPTransMode ctrans;
  WFDRTSPProfile cprofile;
  WFDRTSPLowerTrans clowertrans;
  guint32 crtp_port0;
  guint32 crtp_port1;
  gint tcpsock;
  gchar *sessionID;
  WFDHDCPProtection hdcp_version;
  guint32 hdcp_tcpport;
  gboolean hdcp_support;
#ifdef STANDBY_RESUME_CAPABILITY
  gboolean standby_resume_capability_support;
#endif
  guint cseq;
  gboolean keep_alive_flag;
  GMutex *keep_alive_lock;

};

struct _GstRTSPClientClass {
  GObjectClass  parent_class;

  /* signals */
  void     (*closed)        (GstRTSPClient *client);
  void     (*on_error)      (GstRTSPClient *client);
};

GType                 gst_rtsp_client_get_type          (void);

GstRTSPClient *       gst_rtsp_client_new               (void);

void                  gst_rtsp_client_set_server        (GstRTSPClient * client, GstRTSPServer * server);
GstRTSPServer *       gst_rtsp_client_get_server        (GstRTSPClient * client);

void                  gst_rtsp_client_set_session_pool  (GstRTSPClient *client,
                                                         GstRTSPSessionPool *pool);
GstRTSPSessionPool *  gst_rtsp_client_get_session_pool  (GstRTSPClient *client);

void                  gst_rtsp_client_set_media_mapping (GstRTSPClient *client,
                                                         GstRTSPMediaMapping *mapping);
GstRTSPMediaMapping * gst_rtsp_client_get_media_mapping (GstRTSPClient *client);

void                  gst_rtsp_client_set_auth          (GstRTSPClient *client, GstRTSPAuth *auth);
GstRTSPAuth *         gst_rtsp_client_get_auth          (GstRTSPClient *client);

#if 0
gboolean              gst_rtsp_client_accept            (GstRTSPClient *client, int sock);
#else
gboolean
gst_rtsp_client_accept (GstRTSPClient * client, GIOChannel * channel, GSource *source);
#endif
gboolean gst_rtsp_client_negotiate (GstRTSPClient * client);
gboolean gst_rtsp_client_start (GstRTSPClient * client);
gboolean gst_rtsp_client_pause (GstRTSPClient * client);
gboolean gst_rtsp_client_resume (GstRTSPClient * client);
gboolean gst_rtsp_client_standby (GstRTSPClient * client);
gboolean gst_rtsp_client_stop (GstRTSPClient * client);
void gst_rtsp_client_set_params (GstRTSPClient *client, int videosrc_type, gint session_mode, int videobitrate, gint mtu_size, gchar *infile);

G_END_DECLS

#endif /* __GST_RTSP_CLIENT_H__ */
