/* GStreamer
 * Copyright (C) 2008 Wim Taymans <wim.taymans at gmail.com>

* Copyright (c) 2012, 2013 Samsung Electronics Co., Ltd.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *

* * Modifications by Samsung Electronics Co., Ltd.

* 1. Applied Miracast WFD Server function
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>

#include "rtsp-client.h"
#include "rtsp-sdp.h"
#include "rtsp-params.h"

#include <exynos/exynos_drm.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xmd.h>
#include <dri2.h>
#include <utilX.h>

#define DEFAULT_RTSP_TIMEOUT	60
#define DRM_EXYNOS_VIDI_ON  1
#define DRM_EXYNOS_VIDI_OFF  0

//#define WFD_PAD_PROBE

#ifdef WFD_PAD_PROBE
gulong ahs_appsrc_cb_probe_id = 0;
FILE *f;
#endif

/* temporary multicast address until it's configurable somewhere */
#define MCAST_ADDRESS "224.2.0.1"

#define DEFAULT_WFD_MTU_SIZE 1400

static GMutex *tunnels_lock;
static GHashTable *tunnels;

enum
{
  PROP_0,
  PROP_SESSION_POOL,
  PROP_MEDIA_MAPPING,
  PROP_LAST
};

enum
{
  SIGNAL_CLOSED,
  SIGNAL_ERROR,
  SIGNAL_LAST
};

GST_DEBUG_CATEGORY_STATIC (rtsp_client_debug);
#define GST_CAT_DEFAULT rtsp_client_debug

static guint gst_rtsp_client_signals[SIGNAL_LAST] = { 0 };

static void gst_rtsp_client_get_property (GObject * object, guint propid,
    GValue * value, GParamSpec * pspec);
static void gst_rtsp_client_set_property (GObject * object, guint propid,
    const GValue * value, GParamSpec * pspec);
static void gst_rtsp_client_finalize (GObject * obj);

static void client_session_finalized (GstRTSPClient * client,
    GstRTSPSession * session);
static void unlink_session_streams (GstRTSPClient * client,
    GstRTSPSession * session, GstRTSPSessionMedia * media);
static void
set_edid_info(gchar *edid_info, gboolean plug);

static gboolean
gst_rtsp_client_parse_methods (GstRTSPClient * client, GstRTSPMessage * response);
gboolean
keep_alive_condition(gpointer *userdata);
gboolean
gst_rtsp_client_sending_m16_message (GstRTSPClient * client);

#ifdef WFD_PAD_PROBE
static gboolean
gst_dump_data (GstPad * pad, GstMiniObject * obj, gpointer u_data);
#endif
static void gst_rtsp_client_create_srcbin (GstRTSPClient * client);

G_DEFINE_TYPE (GstRTSPClient, gst_rtsp_client, G_TYPE_OBJECT);

static const char fake_edid_info[] = {
    0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x4c, 0x2d, 0x05, 0x05,
    0x00, 0x00, 0x00, 0x00, 0x30, 0x12, 0x01, 0x03, 0x80, 0x10, 0x09, 0x78,
    0x0a, 0xee, 0x91, 0xa3, 0x54, 0x4c, 0x99, 0x26, 0x0f, 0x50, 0x54, 0xbd,
    0xee, 0x00, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
    0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x66, 0x21, 0x50, 0xb0, 0x51, 0x00,
    0x1b, 0x30, 0x40, 0x70, 0x36, 0x00, 0xa0, 0x5a, 0x00, 0x00, 0x00, 0x1e,
    0x01, 0x1d, 0x00, 0x72, 0x51, 0xd0, 0x1e, 0x20, 0x6e, 0x28, 0x55, 0x00,
    0xa0, 0x5a, 0x00, 0x00, 0x00, 0x1e, 0x00, 0x00, 0x00, 0xfd, 0x00, 0x18,
    0x4b, 0x1a, 0x44, 0x17, 0x00, 0x0a, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
    0x00, 0x00, 0x00, 0xfc, 0x00, 0x53, 0x41, 0x4d, 0x53, 0x55, 0x4e, 0x47,
    0x0a, 0x20, 0x20, 0x20, 0x20, 0x20, 0x01, 0xbc, 0x02, 0x03, 0x1e, 0xf1,
    0x46, 0x84, 0x05, 0x03, 0x10, 0x20, 0x22, 0x23, 0x09, 0x07, 0x07, 0x83,
    0x01, 0x00, 0x00, 0xe2, 0x00, 0x0f, 0x67, 0x03, 0x0c, 0x00, 0x10, 0x00,
    0xb8, 0x2d, 0x01, 0x1d, 0x80, 0x18, 0x71, 0x1c, 0x16, 0x20, 0x58, 0x2c,
    0x25, 0x00, 0xa0, 0x5a, 0x00, 0x00, 0x00, 0x9e, 0x8c, 0x0a, 0xd0, 0x8a,
    0x20, 0xe0, 0x2d, 0x10, 0x10, 0x3e, 0x96, 0x00, 0xa0, 0x5a, 0x00, 0x00,
    0x00, 0x18, 0x02, 0x3a, 0x80, 0x18, 0x71, 0x38, 0x2d, 0x40, 0x58, 0x2c,
    0x45, 0x00, 0xa0, 0x5a, 0x00, 0x00, 0x00, 0x1e, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x06
};

static void
gst_rtsp_client_class_init (GstRTSPClientClass * klass)
{
  GObjectClass *gobject_class;

  gobject_class = G_OBJECT_CLASS (klass);

  gobject_class->get_property = gst_rtsp_client_get_property;
  gobject_class->set_property = gst_rtsp_client_set_property;
  gobject_class->finalize = gst_rtsp_client_finalize;

  g_object_class_install_property (gobject_class, PROP_SESSION_POOL,
      g_param_spec_object ("session-pool", "Session Pool",
          "The session pool to use for client session",
          GST_TYPE_RTSP_SESSION_POOL,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, PROP_MEDIA_MAPPING,
      g_param_spec_object ("media-mapping", "Media Mapping",
          "The media mapping to use for client session",
          GST_TYPE_RTSP_MEDIA_MAPPING,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  gst_rtsp_client_signals[SIGNAL_CLOSED] =
      g_signal_new ("closed", G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_LAST,
      G_STRUCT_OFFSET (GstRTSPClientClass, closed), NULL, NULL,
      g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0, G_TYPE_NONE);

  gst_rtsp_client_signals[SIGNAL_ERROR] =
        g_signal_new ("error_noti", G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_LAST,
        G_STRUCT_OFFSET (GstRTSPClientClass, on_error), NULL, NULL,
        g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0, G_TYPE_NONE);

  tunnels =
      g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
  tunnels_lock = g_mutex_new ();

  GST_DEBUG_CATEGORY_INIT (rtsp_client_debug, "rtspclient", 0, "GstRTSPClient");
}

static void
gst_rtsp_client_init (GstRTSPClient * client)
{
  client->state_lock = g_mutex_new ();
  client->keep_alive_lock = g_mutex_new();
  client->keep_alive_flag = TRUE;
  client->cseq = 0;
  client->state_wait = g_cond_new ();
  client->client_state = CLIENT_STATE_UNKNOWN;
  client->videosrc_type = WFD_INI_VSRC_XVIMAGESRC;
  client->session_mode = 0;
  client->infile = NULL;
  client->MTUsize = DEFAULT_WFD_MTU_SIZE;
}

static void
client_unlink_session (GstRTSPClient * client, GstRTSPSession * session)
{
  GList *medias;

  /* unlink all media managed in this session */
  for (medias = session->medias; medias; medias = g_list_next (medias)) {
    unlink_session_streams (client, session,
        (GstRTSPSessionMedia *) medias->data);
  }
}

static void
client_cleanup_sessions (GstRTSPClient * client)
{
  GList *sessions;

  /* remove weak-ref from sessions */
  for (sessions = client->sessions; sessions; sessions = g_list_next (sessions)) {
    GstRTSPSession *session = (GstRTSPSession *) sessions->data;
    g_object_weak_unref (G_OBJECT (session),
        (GWeakNotify) client_session_finalized, client);
    client_unlink_session (client, session);
  }
  g_list_free (client->sessions);
  client->sessions = NULL;
}

/* A client is finalized when the connection is broken */
static void
gst_rtsp_client_finalize (GObject * obj)
{
  GstRTSPClient *client = GST_RTSP_CLIENT (obj);

  GST_INFO ("finalize client %p", client);

  client_cleanup_sessions (client);

  if(client->connection)
  {
    gst_rtsp_connection_close (client->connection);
    gst_rtsp_connection_free(client->connection);
    if (client->watch)
      g_source_destroy ((GSource *) client->watch);
    client->watchid = 0;
    client->watch = NULL;
  }
  if (client->session_pool)
    g_object_unref (client->session_pool);
  if (client->media_mapping)
    g_object_unref (client->media_mapping);
  if (client->auth)
    g_object_unref (client->auth);

  if (client->uri)
    gst_rtsp_url_free (client->uri);
  if (client->media)
    g_object_unref (client->media);
  g_free (client->server_ip);
  set_edid_info(NULL, FALSE);
  if(client->keep_alive_lock)
    g_mutex_free (client->keep_alive_lock);
  client->keep_alive_lock = NULL;
  g_cond_signal (client->state_wait);
  if (client->server)
    g_object_unref (client->server);
  if(client->tcpsock)
    close(client->tcpsock);
  G_OBJECT_CLASS (gst_rtsp_client_parent_class)->finalize (obj);
}

static void
set_edid_info(gchar *edid_info, gboolean plug)
{
  Display *dpy = NULL;
  int drm_fd;
  int screen;
  drm_magic_t magic;
  int eventBase, errorBase;
  int dri2Major, dri2Minor;
  char *driverName, *deviceName;
  struct drm_exynos_vidi_connection vidi;
  GST_INFO("set_edid_info");

  dpy = XOpenDisplay(NULL);
  if (!dpy) {
    GST_ERROR(" Error: fail to open display");
  }
  screen = DefaultScreen(dpy);

  /*DRI2*/
  if (!DRI2QueryExtension (dpy, &eventBase, &errorBase)) {
    GST_ERROR(" Error : DRI2QueryExtension");
  }

  if (!DRI2QueryVersion (dpy, &dri2Major, &dri2Minor)) {
    GST_ERROR("!!Error : DRI2QueryVersion !!");
    return;
  }

  if (!DRI2Connect (dpy, RootWindow(dpy, screen), &driverName, &deviceName)) {
    GST_ERROR( "!!Error : DRI2Connect !!\n");
    return;
  }

  GST_INFO("Open drm device : %s\n", deviceName);

  /* get the drm_fd though opening the deviceName */
  drm_fd = open (deviceName, O_RDWR);
  if (drm_fd < 0) {
    GST_ERROR("!!Error : cannot open drm device (%s)", deviceName);
    return;
  }
  /* get the drm magic */
  drmGetMagic(drm_fd, &magic);
  GST_INFO(">>> drm magic=%d", magic);

  if (!DRI2Authenticate(dpy, RootWindow(dpy, screen), magic)) {
    GST_ERROR("!!Error : DRI2Authenticate !!");
    close(drm_fd);
    return;
  }
  if (plug == TRUE){
    vidi.connection = DRM_EXYNOS_VIDI_ON;
    vidi.extensions = DRM_EXYNOS_VIDI_ON;
    vidi.edid = (uint64_t *)fake_edid_info;
    GST_ERROR(">>>>>>>>>>>>>>>>>>>>set\n");
  } else {
    vidi.connection = DRM_EXYNOS_VIDI_OFF;
    vidi.extensions = DRM_EXYNOS_VIDI_OFF;
    vidi.edid = NULL;
    GST_ERROR(">>>>>>>>>>>>>>>>>>>>unset\n");
  }

  ioctl (drm_fd, DRM_IOCTL_EXYNOS_VIDI_CONNECTION, &vidi);

  close(drm_fd);

  return;
}

static void
gst_rtsp_client_get_property (GObject * object, guint propid,
    GValue * value, GParamSpec * pspec)
{
  GstRTSPClient *client = GST_RTSP_CLIENT (object);

  switch (propid) {
    case PROP_SESSION_POOL:
      g_value_take_object (value, gst_rtsp_client_get_session_pool (client));
      break;
    case PROP_MEDIA_MAPPING:
      g_value_take_object (value, gst_rtsp_client_get_media_mapping (client));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, propid, pspec);
  }
}

static void
gst_rtsp_client_set_property (GObject * object, guint propid,
    const GValue * value, GParamSpec * pspec)
{
  GstRTSPClient *client = GST_RTSP_CLIENT (object);

  switch (propid) {
    case PROP_SESSION_POOL:
      gst_rtsp_client_set_session_pool (client, g_value_get_object (value));
      break;
    case PROP_MEDIA_MAPPING:
      gst_rtsp_client_set_media_mapping (client, g_value_get_object (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, propid, pspec);
  }
}

/**
 * gst_rtsp_client_new:
 *
 * Create a new #GstRTSPClient instance.
 *
 * Returns: a new #GstRTSPClient
 */
GstRTSPClient *
gst_rtsp_client_new (void)
{
  GstRTSPClient *result;

  result = g_object_new (GST_TYPE_RTSP_CLIENT, NULL);

  return result;
}

static void
send_response (GstRTSPClient * client, GstRTSPSession * session,
    GstRTSPMessage * response)
{
#if 0 //Changes for testing with dongle
  gst_rtsp_message_add_header (response, GST_RTSP_HDR_SERVER,
      "GStreamer RTSP server");
#endif
  /* remove any previous header */
  gst_rtsp_message_remove_header (response, GST_RTSP_HDR_SESSION, -1);

  /* add the new session header for new session ids */
  if (session) {
    gchar *str;

    GST_DEBUG ("session is not NULL - %p", session);

    if (session->timeout > 30)
      str = g_strdup_printf ("%s; timeout=%d", session->sessionid, session->timeout);
    else
    {
      str = g_strdup_printf ("%s;timeout=%d", session->sessionid, 30);
    }

    gst_rtsp_message_take_header (response, GST_RTSP_HDR_SESSION, str);
  }

  if (gst_debug_category_get_threshold (rtsp_client_debug) >= GST_LEVEL_LOG) {
    gst_rtsp_message_dump (response);
  }

  gst_rtsp_watch_send_message (client->watch, response, NULL);
  gst_rtsp_message_unset (response);
}


static void
send_request (GstRTSPClient * client, GstRTSPSession * session, GstRTSPMessage * request)
{
  /* remove any previous header */
  gst_rtsp_message_remove_header (request, GST_RTSP_HDR_SESSION, -1);

  /* add the new session header for new session ids */
  if (session) {
    gchar *str;

    GST_DEBUG ("session is not NULL - %p", session);

    if (session->timeout != DEFAULT_RTSP_TIMEOUT)
      str =
          g_strdup_printf ("%s; timeout=%d", session->sessionid,
          session->timeout);
    else
      str = g_strdup (session->sessionid);

    gst_rtsp_message_take_header (request, GST_RTSP_HDR_SESSION, str);
  }

  if (gst_debug_category_get_threshold (rtsp_client_debug) >= GST_LEVEL_LOG) {
    gst_rtsp_message_dump (request);
  }

  gst_rtsp_watch_send_message (client->watch, request, NULL);
  gst_rtsp_message_unset (request);
}

static void
send_generic_response (GstRTSPClient * client, GstRTSPStatusCode code,
    GstRTSPClientState * state)
{
  gst_rtsp_message_init_response (state->response, code,
      gst_rtsp_status_as_text (code), state->request);

  send_response (client, NULL, state->response);
}

static void
handle_unauthorized_request (GstRTSPClient * client, GstRTSPAuth * auth,
    GstRTSPClientState * state)
{
  gst_rtsp_message_init_response (state->response, GST_RTSP_STS_UNAUTHORIZED,
      gst_rtsp_status_as_text (GST_RTSP_STS_UNAUTHORIZED), state->request);

  if (auth) {
    /* and let the authentication manager setup the auth tokens */
    gst_rtsp_auth_setup_auth (auth, client, 0, state);
  }

  send_response (client, state->session, state->response);
}


static gboolean
compare_uri (const GstRTSPUrl * uri1, const GstRTSPUrl * uri2)
{
  if (uri1 == NULL || uri2 == NULL)
    return FALSE;

  if (strcmp (uri1->abspath, uri2->abspath))
    return FALSE;

  return TRUE;
}

/* this function is called to initially find the media for the DESCRIBE request
 * but is cached for when the same client (without breaking the connection) is
 * doing a setup for the exact same url. */
static GstRTSPMedia *
find_media (GstRTSPClient * client, GstRTSPClientState * state)
{
  GstRTSPMediaFactory *factory;
  GstRTSPMedia *media;
  GstRTSPAuth *auth;

  GST_INFO_OBJECT (client, "client->uri = %s & state->uri = %s",
    client->uri, state->uri);

  if (!compare_uri (client->uri, state->uri)) {
    /* remove any previously cached values before we try to construct a new
     * media for uri */
    if (client->uri)
      gst_rtsp_url_free (client->uri);
    client->uri = NULL;
    if (client->media)
      g_object_unref (client->media);
    client->media = NULL;

    if (!client->media_mapping)
      goto no_mapping;

    GST_DEBUG ("Going to find factory for media mapping...");

    /* find the factory for the uri first */
    if (!(factory =
            gst_rtsp_media_mapping_find_factory (client->media_mapping,
                state->uri)))
      goto no_factory;

    state->factory = factory;

    GST_DEBUG ("Going to get media factory auth...");

    /* check if we have access to the factory */
    if ((auth = gst_rtsp_media_factory_get_auth (factory))) {
      if (!gst_rtsp_auth_check (auth, client, 0, state))
        goto not_allowed;

      g_object_unref (auth);
    }

    GST_DEBUG ("Going to do media factory construct...");

    GST_DEBUG ("Setting source bin...");
    gst_rtsp_media_factory_set_srcbin (factory, GST_ELEMENT(client->srcbin->srcbin));

    /* prepare the media and add it to the pipeline */
    if (!(media = gst_rtsp_media_factory_construct (factory, state->uri)))
      goto no_media;

    g_object_unref (factory);

    /* set ipv6 on the media before preparing */
    media->is_ipv6 = client->is_ipv6;
    state->media = media;

    GST_DEBUG ("Going to do prepare media...");

    /* prepare the media */
    if (!(gst_rtsp_media_prepare (media)))
      goto no_prepare;

    GST_DEBUG ("prepare media done...");

    /* now keep track of the uri and the media */
    client->uri = gst_rtsp_url_copy (state->uri);
    client->media = media;
  } else {
    /* we have seen this uri before, used cached media */
    media = client->media;
    state->media = media;
    GST_INFO ("reusing cached media %p", media);
  }

  if (media)
    g_object_ref (media);

  return media;

  /* ERRORS */
no_mapping:
  {
    send_generic_response (client, GST_RTSP_STS_NOT_FOUND, state);
    return NULL;
  }
no_factory:
  {
    send_generic_response (client, GST_RTSP_STS_NOT_FOUND, state);
    return NULL;
  }
not_allowed:
  {
    handle_unauthorized_request (client, auth, state);
    g_object_unref (factory);
    g_object_unref (auth);
    return NULL;
  }
no_media:
  {
    send_generic_response (client, GST_RTSP_STS_SERVICE_UNAVAILABLE, state);
    g_object_unref (factory);
    return NULL;
  }
no_prepare:
  {
    send_generic_response (client, GST_RTSP_STS_SERVICE_UNAVAILABLE, state);
    g_object_unref (media);
    g_object_unref (factory);
    return NULL;
  }
}

static gboolean
do_send_data (GstBuffer * buffer, guint8 channel, GstRTSPClient * client)
{
  GstRTSPMessage message = { 0 };
  guint8 *data;
  guint size;

  gst_rtsp_message_init_data (&message, channel);

  data = GST_BUFFER_DATA (buffer);
  size = GST_BUFFER_SIZE (buffer);
  gst_rtsp_message_take_body (&message, data, size);

  /* FIXME, client->watch could have been finalized here, we need to keep an
   * extra refcount to the watch.  */
  gst_rtsp_watch_send_message (client->watch, &message, NULL);

  gst_rtsp_message_steal_body (&message, &data, &size);
  gst_rtsp_message_unset (&message);

  return TRUE;
}

static gboolean
do_send_data_list (GstBufferList * blist, guint8 channel,
    GstRTSPClient * client)
{
  GstBufferListIterator *it;

  it = gst_buffer_list_iterate (blist);
  while (gst_buffer_list_iterator_next_group (it)) {
    GstBuffer *group = gst_buffer_list_iterator_merge_group (it);

    if (group == NULL)
      continue;

    do_send_data (group, channel, client);
  }
  gst_buffer_list_iterator_free (it);

  return TRUE;
}

static void
link_stream (GstRTSPClient * client, GstRTSPSession * session,
    GstRTSPSessionStream * stream)
{
  GST_DEBUG ("client %p: linking stream %p", client, stream);
  gst_rtsp_session_stream_set_callbacks (stream, (GstRTSPSendFunc) do_send_data,
      (GstRTSPSendFunc) do_send_data, (GstRTSPSendListFunc) do_send_data_list,
      (GstRTSPSendListFunc) do_send_data_list, client, NULL);
  client->streams = g_list_prepend (client->streams, stream);
  /* make sure our session can't expire */
  gst_rtsp_session_prevent_expire (session);
}

static void
unlink_stream (GstRTSPClient * client, GstRTSPSession * session,
    GstRTSPSessionStream * stream)
{
  GST_DEBUG ("client %p: unlinking stream %p", client, stream);
  gst_rtsp_session_stream_set_callbacks (stream, NULL, NULL, NULL, NULL, NULL,
      NULL);
  client->streams = g_list_remove (client->streams, stream);
  /* our session can now expire */
  gst_rtsp_session_allow_expire (session);
}

static void
unlink_session_streams (GstRTSPClient * client, GstRTSPSession * session,
    GstRTSPSessionMedia * media)
{
  guint n_streams, i;

  n_streams = gst_rtsp_media_n_streams (media->media);
  for (i = 0; i < n_streams; i++) {
    GstRTSPSessionStream *sstream;
    GstRTSPTransport *tr;

    /* get the stream as configured in the session */
    sstream = gst_rtsp_session_media_get_stream (media, i);
    /* get the transport, if there is no transport configured, skip this stream */
    if (!(tr = sstream->trans.transport))
      continue;

    if (tr->lower_transport == GST_RTSP_LOWER_TRANS_TCP) {
      /* for TCP, unlink the stream from the TCP connection of the client */
      unlink_stream (client, session, sstream);
    }
  }
}

static void
close_connection (GstRTSPClient * client)
{
  const gchar *tunnelid;

  GST_DEBUG ("client %p: closing connection", client);

  if ((tunnelid = gst_rtsp_connection_get_tunnelid (client->connection))) {
    g_mutex_lock (tunnels_lock);
    /* remove from tunnelids */
    g_hash_table_remove (tunnels, tunnelid);
    g_mutex_unlock (tunnels_lock);
  }

  gst_rtsp_connection_close (client->connection);
  if (client->watchid) {
    g_source_destroy ((GSource *) client->watch);
    client->watchid = 0;
    client->watch = NULL;
  }
}

static gboolean
handle_teardown_request (GstRTSPClient * client, GstRTSPClientState * state)
{
  GstRTSPSession *session;
  GstRTSPSessionMedia *media;
  GstRTSPStatusCode code;

  if (!state->session)
    goto no_session;

  session = state->session;

  /* get a handle to the configuration of the media in the session */
  media = gst_rtsp_session_get_media (session, state->uri);
  if (!media)
    goto not_found;

  state->sessmedia = media;

  /* unlink the all TCP callbacks */
  unlink_session_streams (client, session, media);

  /* remove the session from the watched sessions */
  g_object_weak_unref (G_OBJECT (session),
      (GWeakNotify) client_session_finalized, client);
  client->sessions = g_list_remove (client->sessions, session);

  gst_rtsp_session_media_set_state (media, GST_STATE_PAUSED);
  gst_rtsp_session_media_set_state (media, GST_STATE_NULL);

  /* unmanage the media in the session, returns false if all media session
   * are torn down. */
  if (!gst_rtsp_session_release_media (session, media)) {
    /* remove the session */
    gst_rtsp_session_pool_remove (client->session_pool, session);
  }
  /* construct the response now */
  code = GST_RTSP_STS_OK;
  gst_rtsp_message_init_response (state->response, code,
      gst_rtsp_status_as_text (code), state->request);

  gst_rtsp_message_add_header (state->response, GST_RTSP_HDR_CONNECTION,
      "close");

  send_response (client, session, state->response);

  close_connection (client);

  return TRUE;

  /* ERRORS */
no_session:
  {
    send_generic_response (client, GST_RTSP_STS_SESSION_NOT_FOUND, state);
    return FALSE;
  }
not_found:
  {
    send_generic_response (client, GST_RTSP_STS_NOT_FOUND, state);
    return FALSE;
  }
}
static GstRTSPResult
prepare_src_response (GstRTSPClient *client, GstRTSPMessage *response, GstRTSPMessage *request, GstRTSPMethod method, WFDMessageType message_type)
{
  if(method == GST_RTSP_GET_PARAMETER) {
    //gst_rtsp_message_init_response (response, GST_RTSP_STS_OK, gst_rtsp_status_as_text (GST_RTSP_STS_OK), request);
    return GST_RTSP_EINVAL;
  } else if(method == GST_RTSP_SET_PARAMETER) {
    switch(message_type) {
    case WFD_MESSAGE_10:
    case WFD_MESSAGE_11:
    case WFD_MESSAGE_12:
    case WFD_MESSAGE_13:
    case WFD_MESSAGE_14:
    case WFD_MESSAGE_15:
      gst_rtsp_message_init_response (response, GST_RTSP_STS_OK, gst_rtsp_status_as_text (GST_RTSP_STS_OK), request);
      return GST_RTSP_OK;
    default: return GST_RTSP_EINVAL;
    }
  }
  return GST_RTSP_EINVAL;
}

WFDMessageType get_rtsp_source_config_message_type(WFDMessage *msg, GstRTSPMethod message_type)
{
  if(message_type == GST_RTSP_GET_PARAMETER) {
    return WFD_MESSAGE_UNKNOWN;
  } else if(message_type == GST_RTSP_SET_PARAMETER) {
    if (msg->route) return WFD_MESSAGE_10;
    else if (msg->connector_type) return WFD_MESSAGE_11;
    else if (msg->standby) return WFD_MESSAGE_12;
    else if (msg->idr_request) return WFD_MESSAGE_13;
    else if (msg->uibc_capability) return WFD_MESSAGE_14;
    else if (msg->uibc_setting) return WFD_MESSAGE_15;
    else return WFD_MESSAGE_UNKNOWN;
  }
  return WFD_MESSAGE_UNKNOWN;
}

static gboolean
handle_get_param_request (GstRTSPClient * client, GstRTSPClientState * state)
{
  GstRTSPResult res;
  guint8 *data;
  guint size;
  WFDMessageType message_type;

  res = gst_rtsp_message_get_body (state->request, &data, &size);
  if (res != GST_RTSP_OK)
    goto bad_request;

  if (size == 0) {
    /* no body, keep-alive request */
    send_generic_response (client, GST_RTSP_STS_OK, state);
  } else {
    WFDMessage *msg;
    wfdconfig_message_new(&msg);
    wfdconfig_message_init(msg);
    wfdconfig_message_parse_buffer(data,size,msg);
    //GST_DEBUG("\nM3 response server side message body: %s\n\n\n", wfdconfig_message_as_text(msg));
    message_type = get_rtsp_source_config_message_type (msg, GST_RTSP_GET_PARAMETER);

    res = prepare_src_response(client, state->response, state->request, GST_RTSP_GET_PARAMETER, message_type);
    if (res != GST_RTSP_OK) {
    /* there is a body, handle the params */
      res = gst_rtsp_params_get (client, state);
      if (res != GST_RTSP_OK)
        goto bad_request;
    }
    send_response (client, state->session, state->response);
  }
  return TRUE;

  /* ERRORS */
bad_request:
  {
    send_generic_response (client, GST_RTSP_STS_BAD_REQUEST, state);
    return FALSE;
  }
}

static guint32 gIDRReqcount = 0;
static gboolean
handle_set_param_request (GstRTSPClient * client, GstRTSPClientState * state)
{
  GstRTSPResult res;
  guint8 *data;
  guint size;
  WFDMessageType message_type;
  WFDResult confret;

  res = gst_rtsp_message_get_body (state->request, &data, &size);
  if (res != GST_RTSP_OK)
    goto bad_request;

  if (size == 0) {
    /* no body, keep-alive request */
    send_generic_response (client, GST_RTSP_STS_OK, state);
  } else {
    WFDMessage *msg;
    wfdconfig_message_new(&msg);
    wfdconfig_message_init(msg);
    wfdconfig_message_parse_buffer(data,size,msg);
    //GST_DEBUG("\nM3 response server side message body: %s\n\n\n", wfdconfig_message_as_text(msg));
    message_type = get_rtsp_source_config_message_type (msg, GST_RTSP_SET_PARAMETER);
    switch (message_type) {
      case WFD_MESSAGE_10:
      {
        WFDSinkType sinktype;
        confret = wfdconfig_get_audio_sink_type(msg, &sinktype);
        if(confret == WFD_OK)
          GST_DEBUG("M10 server set param request route to  %s  sink", sinktype?"SECONDARY":"PRIMARY");
      }
      break;
      case WFD_MESSAGE_11:
      {
        WFDConnector connector;
        confret = wfdconfig_get_connector_type(msg, &connector);
        if(confret == WFD_OK)
          GST_DEBUG("M11 server set param request connector type to %d", connector);
      }
      break;
      case WFD_MESSAGE_12:
      {
        gboolean standby_enable;
        confret = wfdconfig_get_standby(msg, &standby_enable);
        if(confret == WFD_OK)
          GST_DEBUG("M12 server set param request STANDBY %s", standby_enable?"ENABLE":"DISABLE");
      }
      break;
      case WFD_MESSAGE_13:
      {
        GST_DEBUG("M13 server set param request for IDR frame");
        g_object_set (client->srcbin->venc, "force-i-frame", 1, NULL);
    //gIDRReqcount++;
      }
      break;
      case WFD_MESSAGE_14:
      {
        guint32 input_category;
        guint32 inp_type;
        WFDHIDCTypePathPair *inp_pair;
        guint32 inp_type_path_count = 0;
        guint32 i = 0;
        guint32 tcp_port;
        confret = wfdconfig_get_uibc_capability(msg, &input_category, &inp_type, inp_pair, &inp_type_path_count, &tcp_port);
        if(confret == WFD_OK) {
          GST_DEBUG("M14 server set param UIBC input category: %s %s", input_category&WFD_UIBC_INPUT_CAT_GENERIC?"GENERIC":"", input_category&WFD_UIBC_INPUT_CAT_HIDC?"HIDC":"");
          GST_DEBUG("UIBC input type: %s %s %s %s %s %s %s %s", inp_type&WFD_UIBC_INPUT_TYPE_KEYBOARD?"KEYBOARD":"", inp_type&WFD_UIBC_INPUT_TYPE_MOUSE?"MOUSE":"",
             inp_type&WFD_UIBC_INPUT_TYPE_SINGLETOUCH?"SINGLETOUCH":"", inp_type&WFD_UIBC_INPUT_TYPE_MULTITOUCH?"MULTITOUCH":"",
             inp_type&WFD_UIBC_INPUT_TYPE_JOYSTICK?"JOYSTICK":"", inp_type&WFD_UIBC_INPUT_TYPE_CAMERA?"CAMERA":"",
             inp_type&WFD_UIBC_INPUT_TYPE_GESTURE?"GESTURE":"", inp_type&WFD_UIBC_INPUT_TYPE_REMOTECONTROL?"REMOTECONTROL":"");
          for(;i<inp_type_path_count;i++) {
            GST_DEBUG("UIBC %d: %s%s%s%s%s%s%s%s %s%s%s%s%s%s", i, inp_pair[i].inp_type&WFD_UIBC_INPUT_TYPE_KEYBOARD?"KEYBOARD":"", inp_pair[i].inp_type&WFD_UIBC_INPUT_TYPE_MOUSE?"MOUSE":"",
              inp_pair[i].inp_type&WFD_UIBC_INPUT_TYPE_SINGLETOUCH?"SINGLETOUCH":"", inp_pair[i].inp_type&WFD_UIBC_INPUT_TYPE_MULTITOUCH?"MULTITOUCH":"",
              inp_pair[i].inp_type&WFD_UIBC_INPUT_TYPE_JOYSTICK?"JOYSTICK":"", inp_pair[i].inp_type&WFD_UIBC_INPUT_TYPE_CAMERA?"CAMERA":"",
              inp_pair[i].inp_type&WFD_UIBC_INPUT_TYPE_GESTURE?"GESTURE":"", inp_pair[i].inp_type&WFD_UIBC_INPUT_TYPE_REMOTECONTROL?"REMOTECONTROL":"",
              inp_pair[i].inp_path&WFD_UIBC_INPUT_PATH_INFRARED?"INFRARED":"", inp_pair[i].inp_path&WFD_UIBC_INPUT_PATH_USB?"USB":"",
              inp_pair[i].inp_path&WFD_UIBC_INPUT_PATH_BT?"BT":"", inp_pair[i].inp_path&WFD_UIBC_INPUT_PATH_ZIGBEE?"ZIGBEE":"",
              inp_pair[i].inp_path&WFD_UIBC_INPUT_PATH_WIFI?"WIFI":"", inp_pair[i].inp_path&WFD_UIBC_INPUT_PATH_NOSP?"NOSP":"");
          }
        }
      }
      break;
      case WFD_MESSAGE_15:
      {
        gboolean uibc_enable;
        confret = wfdconfig_get_uibc_status(msg, &uibc_enable);
        if(confret == WFD_OK)
          GST_DEBUG("M15 server set param request for UIBC %s", uibc_enable?"ENABLE":"DISABLE");
      }
      break;
      default:
      {
        GST_ERROR_OBJECT (client, "Not handling message num - %d", message_type);
        goto bad_request;
      }
      break;
    }
    res = prepare_src_response(client, state->response, state->request, GST_RTSP_SET_PARAMETER, message_type);
    if (res != GST_RTSP_OK) {
      /* there is a body, handle the params */
      res = gst_rtsp_params_set (client, state);
      if (res != GST_RTSP_OK)
        goto bad_request;
    }
    send_response (client, state->session, state->response);
  }

  return TRUE;

  /* ERRORS */
bad_request:
  {
    send_generic_response (client, GST_RTSP_STS_BAD_REQUEST, state);
    return TRUE;
  }
}

static gboolean
handle_pause_request (GstRTSPClient * client, GstRTSPClientState * state)
{
  GstRTSPSession *session;
  GstRTSPSessionMedia *media;
  GstRTSPStatusCode code;

  if (!(session = state->session))
    goto no_session;

  /* get a handle to the configuration of the media in the session */
  media = gst_rtsp_session_get_media (session, state->uri);
  if (!media)
    goto not_found;

  state->sessmedia = media;

  /* the session state must be playing or recording */
  if (media->state != GST_RTSP_STATE_PLAYING &&
      media->state != GST_RTSP_STATE_RECORDING)
    goto invalid_state;

  /* unlink the all TCP callbacks */
  unlink_session_streams (client, session, media);

  /* then pause sending */
  gst_rtsp_session_media_set_state (media, GST_STATE_PAUSED);

  /* construct the response now */
  code = GST_RTSP_STS_OK;
  gst_rtsp_message_init_response (state->response, code,
      gst_rtsp_status_as_text (code), state->request);

  send_response (client, session, state->response);

  /* the state is now READY */
  media->state = GST_RTSP_STATE_READY;

  return TRUE;

  /* ERRORS */
no_session:
  {
    send_generic_response (client, GST_RTSP_STS_SESSION_NOT_FOUND, state);
    return FALSE;
  }
not_found:
  {
    send_generic_response (client, GST_RTSP_STS_NOT_FOUND, state);
    return FALSE;
  }
invalid_state:
  {
    send_generic_response (client, GST_RTSP_STS_METHOD_NOT_VALID_IN_THIS_STATE,
        state);
    return FALSE;
  }
}

static gboolean
handle_play_request (GstRTSPClient * client, GstRTSPClientState * state)
{
  GstRTSPSession *session;
  GstRTSPSessionMedia *media;
  GstRTSPStatusCode code;
  GString *rtpinfo;
  guint n_streams, i, infocount;
  guint timestamp, seqnum;
  gchar *str;
  GstRTSPTimeRange *range;
  GstRTSPResult res;

  if (!(session = state->session))
    goto no_session;

  /* get a handle to the configuration of the media in the session */
  media = gst_rtsp_session_get_media (session, state->uri);
  if (!media)
    goto not_found;

  state->sessmedia = media;

  /* the session state must be playing or ready */
  if (media->state != GST_RTSP_STATE_PLAYING &&
      media->state != GST_RTSP_STATE_READY)
    goto invalid_state;

  /* parse the range header if we have one */
  res =
      gst_rtsp_message_get_header (state->request, GST_RTSP_HDR_RANGE, &str, 0);
  if (res == GST_RTSP_OK) {
    if (gst_rtsp_range_parse (str, &range) == GST_RTSP_OK) {
      /* we have a range, seek to the position */
      gst_rtsp_media_seek (media->media, range);
      gst_rtsp_range_free (range);
    }
  }

  /* grab RTPInfo from the payloaders now */
  rtpinfo = g_string_new ("");

  n_streams = gst_rtsp_media_n_streams (media->media);
  for (i = 0, infocount = 0; i < n_streams; i++) {
    GstRTSPSessionStream *sstream;
    GstRTSPMediaStream *stream;
    GstRTSPTransport *tr;
    GObjectClass *payobjclass;
    gchar *uristr;

    /* get the stream as configured in the session */
    sstream = gst_rtsp_session_media_get_stream (media, i);
    /* get the transport, if there is no transport configured, skip this stream */
    if (!(tr = sstream->trans.transport)) {
      GST_INFO ("stream %d is not configured", i);
      continue;
    }

    if (tr->lower_transport == GST_RTSP_LOWER_TRANS_TCP) {
      /* for TCP, link the stream to the TCP connection of the client */
      link_stream (client, session, sstream);
    }

    stream = sstream->media_stream;

    payobjclass = G_OBJECT_GET_CLASS (stream->payloader);

    if (g_object_class_find_property (payobjclass, "seqnum") &&
        g_object_class_find_property (payobjclass, "timestamp")) {
      GObject *payobj;

      payobj = G_OBJECT (stream->payloader);

      /* only add RTP-Info for streams with seqnum and timestamp */
      g_object_get (payobj, "seqnum", &seqnum, "timestamp", &timestamp, NULL);

      if (infocount > 0)
        g_string_append (rtpinfo, ", ");

      uristr = gst_rtsp_url_get_request_uri (state->uri);
      g_string_append_printf (rtpinfo, "url=%s;seq=%u;rtptime=%u",
          uristr,seqnum, timestamp);
      g_free (uristr);

      infocount++;
    } else {
      GST_WARNING ("RTP-Info cannot be determined for stream %d", i);
    }
  }

  /* construct the response now */
  code = GST_RTSP_STS_OK;
  gst_rtsp_message_init_response (state->response, code,
  gst_rtsp_status_as_text (code), state->request);
  send_response (client, NULL, state->response);

  /* start playing after sending the request */
  GST_INFO("set media to GST_STATE_PLAYING");
  gst_rtsp_session_media_set_state (media, GST_STATE_PLAYING);

  media->state = GST_RTSP_STATE_PLAYING;

  return TRUE;

  /* ERRORS */
no_session:
  {
    send_generic_response (client, GST_RTSP_STS_SESSION_NOT_FOUND, state);
    return FALSE;
  }
not_found:
  {
    send_generic_response (client, GST_RTSP_STS_NOT_FOUND, state);
    return FALSE;
  }
invalid_state:
  {
    send_generic_response (client, GST_RTSP_STS_METHOD_NOT_VALID_IN_THIS_STATE,
        state);
    return FALSE;
  }
}

static void
do_keepalive (GstRTSPSession * session)
{
  GST_INFO ("keep session %p alive", session);
  gst_rtsp_session_touch (session);
}

static gboolean
handle_setup_request (GstRTSPClient * client, GstRTSPClientState * state)
{
  GstRTSPResult res;
  GstRTSPUrl *uri;
  gchar *transport;
  gchar *user_agent; //Changes for testing with dongle
  gchar **transports;
  gboolean have_transport;
  GstRTSPTransport *ct, *st;
  gint i;
  GstRTSPLowerTrans supported;
  GstRTSPStatusCode code;
  GstRTSPSession *session;
  GstRTSPSessionStream *stream;
  gchar *trans_str, *pos;
  guint streamid;
  GstRTSPSessionMedia *media;
  GstRTSPUrl *url;

  uri = state->uri;

  /* the uri contains the stream number we added in the SDP config, which is
   * always /stream=%d so we need to strip that off
   * parse the stream we need to configure, look for the stream in the abspath
   * first and then in the query. */
  if (uri->abspath == NULL || !(pos = strstr (uri->abspath, "/streamid="))) {
    if (uri->query == NULL || !(pos = strstr (uri->query, "/streamid=")))
    {
      GST_ERROR ("bad request...");
      goto bad_request;
    }
  }

  /* we can mofify the parse uri in place */
  *pos = '\0';

  pos += strlen ("/streamid=");
  if (sscanf (pos, "%u", &streamid) != 1)
    goto bad_request;

  res = gst_rtsp_message_get_header (state->request, GST_RTSP_HDR_USER_AGENT,
      &user_agent, 0);  //Changes for testing with dongle

  /* parse the transport */
  res =
      gst_rtsp_message_get_header (state->request, GST_RTSP_HDR_TRANSPORT,
      &transport, 0);
  if (res != GST_RTSP_OK)
    goto no_transport;

  transports = g_strsplit (transport, ",", 0);
  gst_rtsp_transport_new (&ct);

  /* init transports */
  have_transport = FALSE;
  gst_rtsp_transport_init (ct);

  /* our supported transports */
  supported = GST_RTSP_LOWER_TRANS_UDP |
      GST_RTSP_LOWER_TRANS_UDP_MCAST | GST_RTSP_LOWER_TRANS_TCP;

  /* loop through the transports, try to parse */
  for (i = 0; transports[i]; i++) {
    res = gst_rtsp_transport_parse (transports[i], ct);
    if (res != GST_RTSP_OK) {
      /* no valid transport, search some more */
      GST_WARNING ("could not parse transport %s", transports[i]);
      goto next;
    }

    /* we have a transport, see if it's RTP/AVP */
    if (ct->trans != GST_RTSP_TRANS_RTP || ct->profile != GST_RTSP_PROFILE_AVP) {
      GST_WARNING ("invalid transport %s", transports[i]);
      goto next;
    }

    if (!(ct->lower_transport & supported)) {
      GST_WARNING ("unsupported transport %s", transports[i]);
      goto next;
    }

    /* we have a valid transport */
    GST_INFO ("found valid transport %s", transports[i]);
    have_transport = TRUE;
    break;

  next:
    gst_rtsp_transport_init (ct);
  }
  g_strfreev (transports);

  /* we have not found anything usable, error out */
  if (!have_transport)
    goto unsupported_transports;

  if (client->session_pool == NULL)
    goto no_pool;

  /* we have a valid transport now, set the destination of the client. */
  g_free (ct->destination);
  if (ct->lower_transport == GST_RTSP_LOWER_TRANS_UDP_MCAST) {
    ct->destination = g_strdup (MCAST_ADDRESS);
    ct->port.min = client->crtp_port0; //Changes for testing with dongle
    ct->port.max = -1; //Changes for testing with dongle
  } else {
    url = gst_rtsp_connection_get_url (client->connection);
    ct->destination = g_strdup (url->host);
    ct->client_port.min = client->crtp_port0; //Changes for testing with dongle
    ct->client_port.max = -1; //Changes for testing with dongle
  }

  session = state->session;

  if (session) {
    g_object_ref (session);
    /* get a handle to the configuration of the media in the session, this can
     * return NULL if this is a new url to manage in this session. */
    media = gst_rtsp_session_get_media (session, uri);
  } else {
    /* create a session if this fails we probably reached our session limit or
     * something. */
    if (!(session = gst_rtsp_session_pool_create (client->session_pool)))
      goto service_unavailable;
    session->timeout = DEFAULT_RTSP_TIMEOUT;
    state->session = session;
    client->sessionid = g_strdup(session->sessionid);
    /* we need a new media configuration in this session */
    media = NULL;
  }

  /* we have no media, find one and manage it */
  if (media == NULL) {
    GstRTSPMedia *m;

    // TODO: need to decide on arguments...
    gst_rtsp_client_create_srcbin (client);
    if (NULL == client->srcbin) {
      GST_ERROR_OBJECT (client, "Failed to create src bin...");
      goto no_stream;
    }

    /* get a handle to the configuration of the media in the session */
    if ((m = find_media (client, state))) {
      /* manage the media in our session now */
      media = gst_rtsp_session_manage_media (session, uri, m);
    }
  }

  /* if we stil have no media, error */
  if (media == NULL)
    goto not_found;

  state->sessmedia = media;

  /* fix the transports */
  if (ct->lower_transport & GST_RTSP_LOWER_TRANS_TCP) {
    /* check if the client selected channels for TCP */
    if (ct->interleaved.min == -1 || ct->interleaved.max == -1) {
      gst_rtsp_session_media_alloc_channels (media, &ct->interleaved);
    }
  }

  /* get a handle to the stream in the media */
  if (!(stream = gst_rtsp_session_media_get_stream (media, streamid)))
    goto no_stream;

  st = gst_rtsp_session_stream_set_transport (stream, ct);

  /* configure keepalive for this transport */
  gst_rtsp_session_stream_set_keepalive (stream,
      (GstRTSPKeepAliveFunc) do_keepalive, session, NULL);

  /* serialize the server transport */
  trans_str = gst_rtsp_transport_as_text (st);
  gst_rtsp_transport_free (st);

  /* construct the response now */
  code = GST_RTSP_STS_OK;
  gst_rtsp_message_init_response (state->response, code,
      gst_rtsp_status_as_text (code), state->request);

  gst_rtsp_message_add_header (state->response, GST_RTSP_HDR_TRANSPORT,
      trans_str);
  g_free (trans_str);
/*
  GST_ERROR ("user agent %s", user_agent);
  if(user_agent) {
    gst_rtsp_message_add_header (state->response, GST_RTSP_HDR_USER_AGENT, user_agent);
    g_free (user_agent);
  }*/ //Changes for testing with dongle
  //gst_rtsp_message_add_header (state->response, GST_RTSP_HDR_CONTENT_TYPE, "text/parameters");
  //textparam = g_strdup ("wfd_presentation_URL: rtsp://192.168.16.1/wfd1.0/streamid=0 none");
  //gst_rtsp_message_add_header (state->response, GST_RTSP_HDR_CONTENT_LENGTH, g_strdup_printf ("%d",strlen(textparam)));
    /* adding data field */
   //gst_rtsp_message_set_body (state->response, textparam, strlen(textparam));

  send_response (client, session, state->response);

  /* update the state */
  switch (media->state) {
    case GST_RTSP_STATE_PLAYING:
    case GST_RTSP_STATE_RECORDING:
    case GST_RTSP_STATE_READY:
      /* no state change */
      break;
    default:
      media->state = GST_RTSP_STATE_READY;
      break;
  }
  g_object_unref (session);

  return TRUE;

  /* ERRORS */
bad_request:
  {
    send_generic_response (client, GST_RTSP_STS_BAD_REQUEST, state);
    return FALSE;
  }
not_found:
  {
    send_generic_response (client, GST_RTSP_STS_NOT_FOUND, state);
    g_object_unref (session);
    return FALSE;
  }
no_stream:
  {
    send_generic_response (client, GST_RTSP_STS_NOT_FOUND, state);
    g_object_unref (media);
    g_object_unref (session);
    return FALSE;
  }
no_transport:
  {
    send_generic_response (client, GST_RTSP_STS_UNSUPPORTED_TRANSPORT, state);
    return FALSE;
  }
unsupported_transports:
  {
    send_generic_response (client, GST_RTSP_STS_UNSUPPORTED_TRANSPORT, state);
    gst_rtsp_transport_free (ct);
    return FALSE;
  }
no_pool:
  {
    send_generic_response (client, GST_RTSP_STS_SERVICE_UNAVAILABLE, state);
    return FALSE;
  }
service_unavailable:
  {
    send_generic_response (client, GST_RTSP_STS_SERVICE_UNAVAILABLE, state);
    return FALSE;
  }
}

static GstSDPMessage *
create_sdp (GstRTSPClient * client, GstRTSPMedia * media)
{
  GstSDPMessage *sdp;
  GstSDPInfo info;
  const gchar *proto;

  gst_sdp_message_new (&sdp);

  /* some standard things first */
  gst_sdp_message_set_version (sdp, "0");

  if (client->is_ipv6)
    proto = "IP6";
  else
    proto = "IP4";

  gst_sdp_message_set_origin (sdp, "-", "1188340656180883", "1", "IN", proto,
      client->server_ip);

  gst_sdp_message_set_session_name (sdp, "Session streamed with GStreamer");
  gst_sdp_message_set_information (sdp, "rtsp-server");
  gst_sdp_message_add_time (sdp, "0", "0", NULL);
  gst_sdp_message_add_attribute (sdp, "tool", "GStreamer");
  gst_sdp_message_add_attribute (sdp, "type", "broadcast");
  gst_sdp_message_add_attribute (sdp, "control", "*");

  info.server_proto = proto;
  if (media->protocols & GST_RTSP_LOWER_TRANS_UDP_MCAST)
    info.server_ip = MCAST_ADDRESS;
  else
    info.server_ip = client->server_ip;

  /* create an SDP for the media object */
  if (!gst_rtsp_sdp_from_media (sdp, &info, media))
    goto no_sdp;

  return sdp;

  /* ERRORS */
no_sdp:
  {
    gst_sdp_message_free (sdp);
    return NULL;
  }
}

/* for the describe we must generate an SDP */
static gboolean
handle_describe_request (GstRTSPClient * client, GstRTSPClientState * state)
{
  GstRTSPResult res;
  GstSDPMessage *sdp;
  guint i, str_len;
  gchar *str, *content_base;
  GstRTSPMedia *media;

  /* check what kind of format is accepted, we don't really do anything with it
   * and always return SDP for now. */
  for (i = 0; ;i++) {
    gchar *accept;
    res =
        gst_rtsp_message_get_header (state->request, GST_RTSP_HDR_ACCEPT,
        &accept, i);
    if (res == GST_RTSP_ENOTIMPL)
      break;

    if (g_ascii_strcasecmp (accept, "application/sdp") == 0)
      break;
  }

  /* find the media object for the uri */
  if (!(media = find_media (client, state)))
    goto no_media;

  /* create an SDP for the media object on this client */
  if (!(sdp = create_sdp (client, media)))
    goto no_sdp;

  g_object_unref (media);

  gst_rtsp_message_init_response (state->response, GST_RTSP_STS_OK,
      gst_rtsp_status_as_text (GST_RTSP_STS_OK), state->request);

  gst_rtsp_message_add_header (state->response, GST_RTSP_HDR_CONTENT_TYPE,
      "application/sdp");

  /* content base for some clients that might screw up creating the setup uri */
  str = gst_rtsp_url_get_request_uri (state->uri);
  str_len = strlen (str);

  /* check for trailing '/' and append one */
  if (str[str_len - 1] != '/') {
    content_base = g_malloc (str_len + 2);
    memcpy (content_base, str, str_len);
    content_base[str_len] = '/';
    content_base[str_len + 1] = '\0';
    g_free (str);
  } else {
    content_base = str;
  }

  GST_INFO ("adding content-base: %s", content_base);

  gst_rtsp_message_add_header (state->response, GST_RTSP_HDR_CONTENT_BASE,
      content_base);
  g_free (content_base);

  /* add SDP to the response body */
  str = gst_sdp_message_as_text (sdp);
  gst_rtsp_message_take_body (state->response, (guint8 *) str, strlen (str));
  gst_sdp_message_free (sdp);

  send_response (client, state->session, state->response);

  return TRUE;

  /* ERRORS */
no_media:
  {
    /* error reply is already sent */
    return FALSE;
  }
no_sdp:
  {
    send_generic_response (client, GST_RTSP_STS_SERVICE_UNAVAILABLE, state);
    g_object_unref (media);
    return FALSE;
  }
}

static gboolean
handle_options_request (GstRTSPClient * client, GstRTSPClientState * state)
{
  GstRTSPMethod options;
  gchar *str;
  GstRTSPMessage request = { 0 };
  GstRTSPMessage response = { 0 };
  GstRTSPResult res = GST_RTSP_OK;
  GstRTSPUrl *url = NULL;
  gchar *url_str = NULL;
  WFDMessage *msg3res;
  WFDAudioFormats caCodec = WFD_AUDIO_UNKNOWN;
  WFDAudioFreq cFreq = WFD_FREQ_UNKNOWN;
  WFDAudioChannels cChanels = WFD_CHANNEL_UNKNOWN;
  guint cBitwidth = 0;
  guint caLatency = 0;
  WFDVideoCodecs cvCodec = WFD_VIDEO_UNKNOWN;
  WFDVideoNativeResolution cNative = WFD_VIDEO_CEA_RESOLUTION;
  guint64 cNativeResolution = 0;
  WFDVideoCEAResolution cCEAResolution = WFD_CEA_UNKNOWN;
  WFDVideoVESAResolution cVESAResolution = WFD_VESA_UNKNOWN;
  WFDVideoHHResolution cHHResolution = WFD_HH_UNKNOWN;
  WFDVideoH264Profile cProfile = WFD_H264_UNKNOWN_PROFILE;
  WFDVideoH264Level cLevel = WFD_H264_LEVEL_UNKNOWN;
  guint32 cMaxHeight = 0;
  guint32 cMaxWidth = 0;
  guint32 cmin_slice_size = 0;
  guint32 cslice_enc_params = 0;
  guint cframe_rate_control = 0;
  guint cvLatency = 0;
  WFDRTSPTransMode ctrans = WFD_RTSP_TRANS_UNKNOWN;
  WFDRTSPProfile cprofile = WFD_RTSP_PROFILE_UNKNOWN;
  WFDRTSPLowerTrans clowertrans = WFD_RTSP_LOWER_TRANS_UNKNOWN;
  guint32 crtp_port0 = 0;
  guint32 crtp_port1 = 0;
  gchar *tmp = NULL;

  options = GST_RTSP_OPTIONS |
      GST_RTSP_PAUSE |
      GST_RTSP_PLAY |
      GST_RTSP_SETUP |
      GST_RTSP_GET_PARAMETER | GST_RTSP_SET_PARAMETER | GST_RTSP_TEARDOWN;

  str = gst_rtsp_options_as_text (options);

  /*append WFD specific method */
  //g_string_append (str, "org.wfa.wfd1.0, ");
  tmp = g_strdup (", org.wfa.wfd1.0");
  g_strlcat (str, tmp, strlen (tmp) + strlen (str) + 1);

  gst_rtsp_message_init_response (state->response, GST_RTSP_STS_OK,
      gst_rtsp_status_as_text (GST_RTSP_STS_OK), state->request);

  gst_rtsp_message_add_header (state->response, GST_RTSP_HDR_PUBLIC, str);
  g_free (str);

  send_response (client, state->session, state->response);

  GST_DEBUG_OBJECT (client, "Sent OPTIONS Response...");
 
  ////////////////////////////////////
  ////    GET_PARAMETER Msg (M3)    /////
  ///////////////////////////////////

  url = gst_rtsp_connection_get_url (client->connection);
  url_str = gst_rtsp_url_get_request_uri (url);

  res = gst_rtsp_message_init_request (&request, GST_RTSP_GET_PARAMETER, url_str);
  if (res < 0)
    goto create_request_failed;

  /* add content type */
  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_CONTENT_TYPE, "text/parameters");

  {
   gchar *msg;
   guint msglen = 0;
   GString *msglength;
   WFDMessage *msg3;
   wfdconfig_message_new(&msg3);
   wfdconfig_message_init(msg3);
   wfdconfig_set_supported_audio_format(msg3, WFD_AUDIO_UNKNOWN, WFD_FREQ_UNKNOWN,WFD_CHANNEL_UNKNOWN, 0, 0);
   wfdconfig_set_supported_video_format(msg3, WFD_VIDEO_UNKNOWN, WFD_VIDEO_CEA_RESOLUTION, WFD_CEA_UNKNOWN,
      WFD_CEA_UNKNOWN, WFD_VESA_UNKNOWN,WFD_HH_UNKNOWN, WFD_H264_UNKNOWN_PROFILE,
      WFD_H264_LEVEL_UNKNOWN,0,0, 0, 0, 0, 0);
   wfdconfig_set_prefered_RTP_ports(msg3, WFD_RTSP_TRANS_UNKNOWN, WFD_RTSP_PROFILE_UNKNOWN,
      WFD_RTSP_LOWER_TRANS_UNKNOWN, 0, 0);

   g_print("===================================M3 Message server side============================\n");
   wfdconfig_message_dump(msg3);
   g_print("=====================================================================================\n");
   msg = wfdconfig_message_as_text(msg3);
   msglen = strlen(msg);
   msglength = g_string_new ("");
   g_string_append_printf (msglength,"%d",msglen);
   GST_DEBUG("M3 server side message body: %s\n\n\n", msg);
      /* add content-length type */
   gst_rtsp_message_add_header (&request, GST_RTSP_HDR_CONTENT_LENGTH, g_string_free (msglength, FALSE));
      /* adding data field */
   gst_rtsp_message_set_body (&request, (guint8*)msg, msglen);

   wfdconfig_message_free(msg3);
  }

  GST_DEBUG ("send GET_PARAM (M3)...");

  // TODO: need to add session i.e. 2nd variable
  send_request (client, NULL, &request);

  res = gst_rtsp_connection_receive (client->connection, &response, NULL);
  if (GST_RTSP_OK != res) {
    goto receive_error;
  }

  if (gst_debug_category_get_threshold (rtsp_client_debug) >= GST_LEVEL_LOG) {
    gst_rtsp_message_dump (&response);
  }

  {
  gchar *data = NULL;
  guint size=0;
  gst_rtsp_message_get_body (&response, (guint8**)&data, &size);

  wfdconfig_message_new(&msg3res);
  wfdconfig_message_init(msg3res);
  wfdconfig_message_parse_buffer((guint8*)data,size,msg3res);
  GST_DEBUG("\nM3 response server side message body: %s\n\n\n", wfdconfig_message_as_text(msg3res));

  wfdconfig_get_supported_audio_format(msg3res, &caCodec, &cFreq, &cChanels, &cBitwidth, &caLatency);
  wfdconfig_get_supported_video_format(msg3res, &cvCodec, &cNative, &cNativeResolution,
                          (guint64 *)&cCEAResolution, (guint64 *)&cVESAResolution, (guint64 *)&cHHResolution,
                          &cProfile, &cLevel, &cvLatency, &cMaxHeight,
                          &cMaxWidth, &cmin_slice_size, &cslice_enc_params, &cframe_rate_control);
  wfdconfig_get_prefered_RTP_ports(msg3res, &ctrans, &cprofile, &clowertrans, &crtp_port0, &crtp_port1);
  g_print("==============================M3 response Message server side========================\n");
  wfdconfig_message_dump(msg3res);
  g_print("=====================================================================================\n");
  }

  ////////////////////////////////////
  ////    SET_PARAMETER Msg (M4)    /////
  ///////////////////////////////////

  memset (&request, 0x00, sizeof (GstRTSPMessage));
  memset (&response, 0x00, sizeof (GstRTSPMessage));

  res = gst_rtsp_message_init_request (&request, GST_RTSP_SET_PARAMETER, url_str);
  if (res < 0)
    goto create_request_failed;
  GST_DEBUG("\nM4 message body server side url : %s\n\n\n", url_str);
  /* add content type */
  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_CONTENT_TYPE, "text/parameters");

  {
   gchar *msg;
   guint msglen = 0;
   GString *msglength;
   WFDMessage *msg4;

   wfdconfig_message_new(&msg4);
   wfdconfig_message_init(msg4);
   wfdconfig_set_prefered_audio_format(msg4, WFD_AUDIO_AAC, WFD_FREQ_48000, WFD_CHANNEL_2, cBitwidth, caLatency);
   wfdconfig_set_prefered_video_format(msg4, WFD_VIDEO_H264,WFD_VIDEO_CEA_RESOLUTION, WFD_CEA_1280x720P30,
        WFD_CEA_1280x720P30, WFD_VESA_1280x768P30,WFD_HH_960x540P30, WFD_H264_BASE_PROFILE,
        WFD_H264_LEVEL_4, cvLatency, cMaxHeight,cMaxWidth, cmin_slice_size, cslice_enc_params, cframe_rate_control);
   wfdconfig_set_prefered_RTP_ports(msg4, WFD_RTSP_TRANS_RTP, WFD_RTSP_PROFILE_AVP, WFD_RTSP_LOWER_TRANS_UDP, crtp_port0, crtp_port1);
   g_print("==================================M4 Message server side=============================\n");
   wfdconfig_message_dump(msg4);
   g_print("=====================================================================================\n");
   msg = wfdconfig_message_as_text(msg4);
   msglen = strlen(msg);
   msglength = g_string_new ("");
   g_string_append_printf (msglength,"%d",msglen);
   GST_DEBUG("\nM4 message body server side : %s\n\n\n", msg);
      /* add content-length type */
   gst_rtsp_message_add_header (&request, GST_RTSP_HDR_CONTENT_LENGTH, g_string_free (msglength, FALSE));
      /* adding data field */
   gst_rtsp_message_set_body (&request, (guint8*)msg, msglen);

   wfdconfig_message_free(msg4);
  }

  GST_DEBUG ("send SET_PARAM (M4)...");

  // TODO: need to add session i.e. 2nd variable
  send_request (client, NULL, &request);

  res = gst_rtsp_connection_receive (client->connection, &response, NULL);
  if (GST_RTSP_OK != res) {
     goto receive_error;
  }

  if (gst_debug_category_get_threshold (rtsp_client_debug) >= GST_LEVEL_LOG) {
    gst_rtsp_message_dump (&response);
  }

  return TRUE;

create_request_failed:
  GST_ERROR_OBJECT (client, "Failed to create server request...reason : %d", res);
  return FALSE;

receive_error:
  GST_ERROR_OBJECT (client, "Failed to receive response .....reason: %d", res);
  return FALSE;
}

/* remove duplicate and trailing '/' */
static void
sanitize_uri (GstRTSPUrl * uri)
{
  gint i, len;
  gchar *s, *d;
  gboolean have_slash, prev_slash;

  s = d = uri->abspath;
  len = strlen (uri->abspath);

  prev_slash = FALSE;

  for (i = 0; i < len; i++) {
    have_slash = s[i] == '/';
    *d = s[i];
    if (!have_slash || !prev_slash)
      d++;
    prev_slash = have_slash;
  }
  len = d - uri->abspath;
  /* don't remove the first slash if that's the only thing left */
  if (len > 1 && *(d - 1) == '/')
    d--;
  *d = '\0';
}

static void
client_session_finalized (GstRTSPClient * client, GstRTSPSession * session)
{
  GST_INFO ("client %p: session %p finished", client, session);

  /* unlink all media managed in this session */
  client_unlink_session (client, session);

  /* remove the session */
  if (!(client->sessions = g_list_remove (client->sessions, session))) {
    GST_INFO ("client %p: all sessions finalized, close the connection",
        client);
    close_connection (client);
  }
}

static void
client_watch_session (GstRTSPClient * client, GstRTSPSession * session)
{
  GList *walk;

  for (walk = client->sessions; walk; walk = g_list_next (walk)) {
    GstRTSPSession *msession = (GstRTSPSession *) walk->data;

    /* we already know about this session */
    if (msession == session)
      return;
  }

  GST_INFO ("watching session %p", session);

  g_object_weak_ref (G_OBJECT (session), (GWeakNotify) client_session_finalized,
      client);
  client->sessions = g_list_prepend (client->sessions, session);
}

static void
handle_request (GstRTSPClient * client, GstRTSPMessage * request)
{
  GstRTSPMethod method;
  const gchar *uristr;
  GstRTSPUrl *uri;
  GstRTSPVersion version;
  GstRTSPResult res;
  GstRTSPSession *session;
  GstRTSPClientState state = { NULL };
  GstRTSPMessage response = { 0 };
  gchar *sessid;

  state.request = request;
  state.response = &response;

  if (gst_debug_category_get_threshold (rtsp_client_debug) >= GST_LEVEL_LOG) {
    gst_rtsp_message_dump (request);
  }

  GST_INFO ("client %p: received a request", client);

  gst_rtsp_message_parse_request (request, &method, &uristr, &version);

  if (version != GST_RTSP_VERSION_1_0) {
    /* we can only handle 1.0 requests */
    send_generic_response (client, GST_RTSP_STS_RTSP_VERSION_NOT_SUPPORTED,
        &state);
    return;
  }
  state.method = method;

  /* we always try to parse the url first */
  if (gst_rtsp_url_parse (uristr, &uri) != GST_RTSP_OK) {
    send_generic_response (client, GST_RTSP_STS_BAD_REQUEST, &state);
    return;
  }

  /* sanitize the uri */
  sanitize_uri (uri);
  state.uri = uri;

  GST_DEBUG ("state -> URI = %s", uri->abspath);

  /* get the session if there is any */
  res = gst_rtsp_message_get_header (request, GST_RTSP_HDR_SESSION, &sessid, 0);
  if (res == GST_RTSP_OK) {
    if (client->session_pool == NULL)
      goto no_pool;

    /* we had a session in the request, find it again */
    if (!(session = gst_rtsp_session_pool_find (client->session_pool, sessid)))
      goto session_not_found;

    GST_DEBUG ("sessid = %s", sessid);

    /* we add the session to the client list of watched sessions. When a session
     * disappears because it times out, we will be notified. If all sessions are
     * gone, we will close the connection */
    client_watch_session (client, session);
    if(client->sessionID) g_free(client->sessionID);
    client->sessionID = g_strdup(sessid);
    } else if(method == GST_RTSP_SET_PARAMETER){
      if (client->session_pool == NULL)
        goto no_pool;
      if (!(session = gst_rtsp_session_pool_find (client->session_pool, client->sessionID)))
        goto session_not_found;
      GST_DEBUG ("sessid = %s", client->sessionID);
      client_watch_session (client, session);
    }
    else session = NULL;

  GST_DEBUG ("session = %p", session);
  state.session = session;


  if(session) {
    GstRTSPSessionMedia *temp = session->medias->data;
    temp->url->abspath = g_strdup(uri->abspath);
  }

  if (client->auth) {
    if (!gst_rtsp_auth_check (client->auth, client, 0, &state))
      goto not_authorized;
  }

  GST_INFO ("Method = %d", method);

  /* now see what is asked and dispatch to a dedicated handler */
  switch (method) {
    case GST_RTSP_OPTIONS:
      //handle_options_request (client, &state);
      break;
    case GST_RTSP_DESCRIBE:
      handle_describe_request (client, &state);
      break;
    case GST_RTSP_SETUP:
      handle_setup_request (client, &state);
      break;
    case GST_RTSP_PLAY:
      handle_play_request (client, &state);
      g_timeout_add((DEFAULT_RTSP_TIMEOUT - 5)*1000, keep_alive_condition, client);
      break;
    case GST_RTSP_PAUSE:
      handle_pause_request (client, &state);
      break;
    case GST_RTSP_TEARDOWN:
      handle_teardown_request (client, &state);
      break;
    case GST_RTSP_SET_PARAMETER:
      handle_set_param_request (client, &state);
      break;
    case GST_RTSP_GET_PARAMETER:
      handle_get_param_request (client, &state);
      break;
    case GST_RTSP_ANNOUNCE:
    case GST_RTSP_RECORD:
    case GST_RTSP_REDIRECT:
      send_generic_response (client, GST_RTSP_STS_NOT_IMPLEMENTED, &state);
      break;
    case GST_RTSP_INVALID:
    default:
      send_generic_response (client, GST_RTSP_STS_BAD_REQUEST, &state);
      break;
  }
  if (session)
    g_object_unref (session);

  gst_rtsp_url_free (uri);
  return;

  /* ERRORS */
no_pool:
  {
    send_generic_response (client, GST_RTSP_STS_SERVICE_UNAVAILABLE, &state);
    return;
  }
session_not_found:
  {
    send_generic_response (client, GST_RTSP_STS_SESSION_NOT_FOUND, &state);
    return;
  }
not_authorized:
  {
    handle_unauthorized_request (client, client->auth, &state);
    return;
  }
}

static void
handle_data (GstRTSPClient * client, GstRTSPMessage * message)
{
  GstRTSPResult res;
  guint8 channel;
  GList *walk;
  guint8 *data;
  guint size;
  GstBuffer *buffer;
  gboolean handled;

  /* find the stream for this message */
  res = gst_rtsp_message_parse_data (message, &channel);
  if (res != GST_RTSP_OK)
    return;

  gst_rtsp_message_steal_body (message, &data, &size);

  buffer = gst_buffer_new ();
  GST_BUFFER_DATA (buffer) = data;
  GST_BUFFER_MALLOCDATA (buffer) = data;
  GST_BUFFER_SIZE (buffer) = size;

  handled = FALSE;
  for (walk = client->streams; walk; walk = g_list_next (walk)) {
    GstRTSPSessionStream *stream = (GstRTSPSessionStream *) walk->data;
    GstRTSPMediaStream *mstream;
    GstRTSPTransport *tr;

    /* get the transport, if there is no transport configured, skip this stream */
    if (!(tr = stream->trans.transport))
      continue;

    /* we also need a media stream */
    if (!(mstream = stream->media_stream))
      continue;

    /* check for TCP transport */
    if (tr->lower_transport == GST_RTSP_LOWER_TRANS_TCP) {
      GST_INFO ("Transport is GST_RTSP_LOWER_TRANS_TCP");
      /* dispatch to the stream based on the port number */
      if (client->crtp_port0) {
        gst_rtsp_media_stream_rtp (mstream, buffer);
        handled = TRUE;
        break;
      } else if (client->crtp_port1) {
        gst_rtsp_media_stream_rtcp (mstream, buffer);
        handled = TRUE;
        break;
      }
    }
  }
  if (!handled)
    gst_buffer_unref (buffer);
}

/**
 * gst_rtsp_client_set_session_pool:
 * @client: a #GstRTSPClient
 * @pool: a #GstRTSPSessionPool
 *
 * Set @pool as the sessionpool for @client which it will use to find
 * or allocate sessions. the sessionpool is usually inherited from the server
 * that created the client but can be overridden later.
 */
void
gst_rtsp_client_set_session_pool (GstRTSPClient * client,
    GstRTSPSessionPool * pool)
{
  GstRTSPSessionPool *old;

  old = client->session_pool;
  if (old != pool) {
    if (pool)
      g_object_ref (pool);
    client->session_pool = pool;
    if (old)
      g_object_unref (old);
  }
}

/**
 * gst_rtsp_client_get_session_pool:
 * @client: a #GstRTSPClient
 *
 * Get the #GstRTSPSessionPool object that @client uses to manage its sessions.
 *
 * Returns: a #GstRTSPSessionPool, unref after usage.
 */
GstRTSPSessionPool *
gst_rtsp_client_get_session_pool (GstRTSPClient * client)
{
  GstRTSPSessionPool *result;

  if ((result = client->session_pool))
    g_object_ref (result);

  return result;
}

/**
 * gst_rtsp_client_set_server:
 * @client: a #GstRTSPClient
 * @server: a #GstRTSPServer
 *
 * Set @server as the server that created @client.
 */
void
gst_rtsp_client_set_server (GstRTSPClient * client, GstRTSPServer * server)
{
  GstRTSPServer *old;

  old = client->server;
  if (old != server) {
    if (server)
      g_object_ref (server);
    client->server = server;
    if (old)
      g_object_unref (old);
  }
}

/**
 * gst_rtsp_client_get_server:
 * @client: a #GstRTSPClient
 *
 * Get the #GstRTSPServer object that @client was created from.
 *
 * Returns: a #GstRTSPServer, unref after usage.
 */
GstRTSPServer *
gst_rtsp_client_get_server (GstRTSPClient * client)
{
  GstRTSPServer *result;

  if ((result = client->server))
    g_object_ref (result);

  return result;
}

/**
 * gst_rtsp_client_set_media_mapping:
 * @client: a #GstRTSPClient
 * @mapping: a #GstRTSPMediaMapping
 *
 * Set @mapping as the media mapping for @client which it will use to map urls
 * to media streams. These mapping is usually inherited from the server that
 * created the client but can be overriden later.
 */
void
gst_rtsp_client_set_media_mapping (GstRTSPClient * client,
    GstRTSPMediaMapping * mapping)
{
  GstRTSPMediaMapping *old;

  old = client->media_mapping;

  if (old != mapping) {
    if (mapping)
      g_object_ref (mapping);
    client->media_mapping = mapping;
    if (old)
      g_object_unref (old);
  }
}

/**
 * gst_rtsp_client_get_media_mapping:
 * @client: a #GstRTSPClient
 *
 * Get the #GstRTSPMediaMapping object that @client uses to manage its sessions.
 *
 * Returns: a #GstRTSPMediaMapping, unref after usage.
 */
GstRTSPMediaMapping *
gst_rtsp_client_get_media_mapping (GstRTSPClient * client)
{
  GstRTSPMediaMapping *result;

  if ((result = client->media_mapping))
    g_object_ref (result);

  return result;
}

/**
 * gst_rtsp_client_set_auth:
 * @client: a #GstRTSPClient
 * @auth: a #GstRTSPAuth
 *
 * configure @auth to be used as the authentication manager of @client.
 */
void
gst_rtsp_client_set_auth (GstRTSPClient * client, GstRTSPAuth * auth)
{
  GstRTSPAuth *old;

  g_return_if_fail (GST_IS_RTSP_CLIENT (client));

  old = client->auth;

  if (old != auth) {
    if (auth)
      g_object_ref (auth);
    client->auth = auth;
    if (old)
      g_object_unref (old);
  }
}


/**
 * gst_rtsp_client_get_auth:
 * @client: a #GstRTSPClient
 *
 * Get the #GstRTSPAuth used as the authentication manager of @client.
 *
 * Returns: the #GstRTSPAuth of @client. g_object_unref() after
 * usage.
 */
GstRTSPAuth *
gst_rtsp_client_get_auth (GstRTSPClient * client)
{
  GstRTSPAuth *result;

  g_return_val_if_fail (GST_IS_RTSP_CLIENT (client), NULL);

  if ((result = client->auth))
    g_object_ref (result);

  return result;
}

static GstRTSPResult
message_received (GstRTSPWatch * watch, GstRTSPMessage * message,
    gpointer user_data)
{
  GstRTSPClient *client = GST_RTSP_CLIENT (user_data);

  GST_INFO_OBJECT (client, "revd message method = %s", 
    message->type == GST_RTSP_MESSAGE_REQUEST ? "request" :
    message->type == GST_RTSP_MESSAGE_RESPONSE ? "reponse" : "data/unknown");
  switch (message->type) {
    case GST_RTSP_MESSAGE_REQUEST:
      handle_request (client, message);
      break;
    case GST_RTSP_MESSAGE_RESPONSE:
      {
        GstRTSPStatusCode code;
        const gchar *uristr;
        GstRTSPVersion version;

        gst_rtsp_message_parse_response (message, &code, &uristr, &version);
        GST_INFO_OBJECT (client, "revd message method = %d", code);
        g_mutex_lock(client->keep_alive_lock);
        client->keep_alive_flag = TRUE;
        g_mutex_unlock(client->keep_alive_lock);
      }
      break;
    case GST_RTSP_MESSAGE_DATA:
      handle_data (client, message);
      break;
    default:
      break;
  }
  return GST_RTSP_OK;
}

static GstRTSPResult
message_sent (GstRTSPWatch * watch, guint cseq, gpointer user_data)
{
  GstRTSPClient *client;

  client = GST_RTSP_CLIENT (user_data);

  return GST_RTSP_OK;
}

static GstRTSPResult
closed (GstRTSPWatch * watch, gpointer user_data)
{
  GstRTSPClient *client = GST_RTSP_CLIENT (user_data);
  const gchar *tunnelid;

  GST_INFO ("client %p: connection closed", client);

  if ((tunnelid = gst_rtsp_connection_get_tunnelid (client->connection))) {
    g_mutex_lock (tunnels_lock);
    /* remove from tunnelids */
    g_hash_table_remove (tunnels, tunnelid);
    g_mutex_unlock (tunnels_lock);
  }

  return GST_RTSP_OK;
}

static GstRTSPResult
error (GstRTSPWatch * watch, GstRTSPResult result, gpointer user_data)
{
  GstRTSPClient *client = GST_RTSP_CLIENT (user_data);
  gchar *str;

  str = gst_rtsp_strresult (result);
  GST_INFO ("client %p: received an error %s", client, str);
  g_free (str);

  return GST_RTSP_OK;
}

static GstRTSPResult
error_full (GstRTSPWatch * watch, GstRTSPResult result,
    GstRTSPMessage * message, guint id, gpointer user_data)
{
  GstRTSPClient *client = GST_RTSP_CLIENT (user_data);
  gchar *str;

  str = gst_rtsp_strresult (result);
  GST_INFO
      ("client %p: received an error %s when handling message %p with id %d",
      client, str, message, id);
  g_free (str);

  return GST_RTSP_OK;
}

static gboolean
remember_tunnel (GstRTSPClient * client)
{
  const gchar *tunnelid;

  /* store client in the pending tunnels */
  tunnelid = gst_rtsp_connection_get_tunnelid (client->connection);
  if (tunnelid == NULL)
    goto no_tunnelid;

  GST_INFO ("client %p: inserting tunnel session %s", client, tunnelid);

  /* we can't have two clients connecting with the same tunnelid */
  g_mutex_lock (tunnels_lock);
  if (g_hash_table_lookup (tunnels, tunnelid))
    goto tunnel_existed;

  g_hash_table_insert (tunnels, g_strdup (tunnelid), g_object_ref (client));
  g_mutex_unlock (tunnels_lock);

  return TRUE;

  /* ERRORS */
no_tunnelid:
  {
    GST_ERROR ("client %p: no tunnelid provided", client);
    return FALSE;
  }
tunnel_existed:
  {
    g_mutex_unlock (tunnels_lock);
    GST_ERROR ("client %p: tunnel session %s already existed", client,
        tunnelid);
    return FALSE;
  }
}

static GstRTSPStatusCode
tunnel_start (GstRTSPWatch * watch, gpointer user_data)
{
  GstRTSPClient *client;

  client = GST_RTSP_CLIENT (user_data);

  GST_INFO ("client %p: tunnel start (connection %p)", client,
      client->connection);

  if (!remember_tunnel (client))
    goto tunnel_error;

  return GST_RTSP_STS_OK;

  /* ERRORS */
tunnel_error:
  {
    GST_ERROR ("client %p: error starting tunnel", client);
    return GST_RTSP_STS_SERVICE_UNAVAILABLE;
  }
}

static GstRTSPResult
tunnel_lost (GstRTSPWatch * watch, gpointer user_data)
{
  GstRTSPClient *client;

  client = GST_RTSP_CLIENT (user_data);

  GST_INFO ("client %p: tunnel lost (connection %p)", client,
      client->connection);

  /* ignore error, it'll only be a problem when the client does a POST again */
  remember_tunnel (client);

  return GST_RTSP_OK;
}

static GstRTSPResult
tunnel_complete (GstRTSPWatch * watch, gpointer user_data)
{
  const gchar *tunnelid;
  GstRTSPClient *client = GST_RTSP_CLIENT (user_data);
  GstRTSPClient *oclient;

  GST_INFO ("client %p: tunnel complete", client);

  /* find previous tunnel */
  tunnelid = gst_rtsp_connection_get_tunnelid (client->connection);
  if (tunnelid == NULL)
    goto no_tunnelid;

  g_mutex_lock (tunnels_lock);
  if (!(oclient = g_hash_table_lookup (tunnels, tunnelid)))
    goto no_tunnel;

  /* remove the old client from the table. ref before because removing it will
   * remove the ref to it. */
  g_object_ref (oclient);
  g_hash_table_remove (tunnels, tunnelid);

  if (oclient->watch == NULL)
    goto tunnel_closed;
  g_mutex_unlock (tunnels_lock);

  GST_INFO ("client %p: found tunnel %p (old %p, new %p)", client, oclient,
      oclient->connection, client->connection);

  /* merge the tunnels into the first client */
  gst_rtsp_connection_do_tunnel (oclient->connection, client->connection);
  gst_rtsp_watch_reset (oclient->watch);
  g_object_unref (oclient);

  /* we don't need this watch anymore */
  g_source_destroy ((GSource *) client->watch);
  client->watchid = 0;
  client->watch = NULL;

  return GST_RTSP_OK;

  /* ERRORS */
no_tunnelid:
  {
    GST_INFO ("client %p: no tunnelid provided", client);
    return GST_RTSP_STS_SERVICE_UNAVAILABLE;
  }
no_tunnel:
  {
    g_mutex_unlock (tunnels_lock);
    GST_INFO ("client %p: tunnel session %s not found", client, tunnelid);
    return GST_RTSP_STS_SERVICE_UNAVAILABLE;
  }
tunnel_closed:
  {
    g_mutex_unlock (tunnels_lock);
    GST_INFO ("client %p: tunnel session %s was closed", client, tunnelid);
    g_object_unref (oclient);
    return GST_RTSP_STS_SERVICE_UNAVAILABLE;
  }
}

static GstRTSPWatchFuncs watch_funcs = {
  message_received,
  message_sent,
  closed,
  error,
  tunnel_start,
  tunnel_complete,
  error_full,
  tunnel_lost
};

static void
client_watch_notify (GstRTSPClient * client)
{
  GST_INFO ("client %p: watch destroyed", client);
  client->watchid = 0;
  client->watch = NULL;
  g_signal_emit (client, gst_rtsp_client_signals[SIGNAL_CLOSED], 0, NULL);
  g_object_unref (client);
}

#if 0
/**
 * gst_rtsp_client_attach:
 * @client: a #GstRTSPClient
 * @channel: a #GIOChannel
 *
 * Accept a new connection for @client on the socket in @channel.
 *
 * This function should be called when the client properties and urls are fully
 * configured and the client is ready to start.
 *
 * Returns: %TRUE if the client could be accepted.
 */
gboolean
gst_rtsp_client_accept (GstRTSPClient * client, int sock)
{
  int fd;
  GstRTSPConnection *conn;
  GstRTSPResult res;
  GSource *source;
  GMainContext *context;
  GstRTSPUrl *url;
  struct sockaddr_storage addr;
  socklen_t addrlen;
  gchar ip[INET6_ADDRSTRLEN];

 // TODO: try about SIGIO or select or poll calls apart from waiting...

  GST_ERROR_OBJECT (client, "Enter accept client...");

  GST_RTSP_CHECK (gst_rtsp_connection_accept (sock, &conn), accept_failed);

  fd = gst_rtsp_connection_get_readfd (conn);

  addrlen = sizeof (addr);
  if (getsockname (fd, (struct sockaddr *) &addr, &addrlen) < 0)
    goto getpeername_failed;

  client->is_ipv6 = addr.ss_family == AF_INET6;

  if (getnameinfo ((struct sockaddr *) &addr, addrlen, ip, sizeof (ip), NULL, 0,
          NI_NUMERICHOST) != 0)
    goto getnameinfo_failed;

  /* keep the original ip that the client connected to */
  g_free (client->server_ip);
  client->server_ip = g_strndup (ip, sizeof (ip));

  GST_INFO ("client %p connected to server ip %s, ipv6 = %d", client,
      client->server_ip, client->is_ipv6);

  url = gst_rtsp_connection_get_url (conn);
  GST_INFO ("added new client %p ip %s:%d", client, url->host, url->port);

  client->connection = conn;

  /* create watch for the connection and attach */
  client->watch = gst_rtsp_watch_new (client->connection, &watch_funcs,
      g_object_ref (client), (GDestroyNotify) client_watch_notify);

  source = client->watch;

  GST_DEBUG (" source = %p", source);

  // naveen: we are not creating source before below call.. so context will be NULL.. which means, we are attaching
  // to main context...

  /* find the context to add the watch */
  if ((source = g_main_current_source ()))
    context = g_source_get_context (source);
  else
    context = NULL;

  GST_DEBUG (" source = %p", source);


  GST_INFO ("attaching to context %p", context);

  client->watchid = gst_rtsp_watch_attach (client->watch, context);

  GST_DEBUG_OBJECT (client, "watch_id = %u", client->watchid);
  gst_rtsp_watch_unref (client->watch);

  return TRUE;

  /* ERRORS */
accept_failed:
  {
    gchar *str = gst_rtsp_strresult (res);

    GST_ERROR ("Could not accept client on server socket %d: %s", sock, str);
    g_free (str);
    return FALSE;
  }
getpeername_failed:
  {
    GST_ERROR ("getpeername failed: %s", g_strerror (errno));
    return FALSE;
  }
getnameinfo_failed:
  {
    GST_ERROR ("getnameinfo failed: %s", g_strerror (errno));
    return FALSE;
  }
}
#else
/**
 * gst_rtsp_client_attach:
 * @client: a #GstRTSPClient
 * @channel: a #GIOChannel
 *
 * Accept a new connection for @client on the socket in @channel.
 *
 * This function should be called when the client properties and urls are fully
 * configured and the client is ready to start.
 *
 * Returns: %TRUE if the client could be accepted.
 */
gboolean
gst_rtsp_client_accept (GstRTSPClient * client, GIOChannel * channel, GSource *source)
{
  int sock, fd;
  GstRTSPConnection *conn;
  GstRTSPResult res;
  GMainContext *context;
  GstRTSPUrl *url;
  struct sockaddr_storage addr;
  socklen_t addrlen;
  gchar ip[INET6_ADDRSTRLEN];

  GST_ERROR_OBJECT (client, "Enter accept client...");

  /* a new client connected. */
  sock = g_io_channel_unix_get_fd (channel);

  GST_INFO_OBJECT (client, "sock fd  = %d", sock);

  GST_RTSP_CHECK (gst_rtsp_connection_accept (sock, &conn), accept_failed);

  fd = gst_rtsp_connection_get_readfd (conn);
  if (fd == -1) {
     res = GST_RTSP_EINVAL;
     goto  accept_failed;
  }

  addrlen = sizeof (addr);
  if (getsockname (fd, (struct sockaddr *) &addr, &addrlen) < 0)
    goto getpeername_failed;

  client->is_ipv6 = addr.ss_family == AF_INET6;

  if (getnameinfo ((struct sockaddr *) &addr, addrlen, ip, sizeof (ip), NULL, 0,
          NI_NUMERICHOST) != 0)
    goto getnameinfo_failed;

  /* keep the original ip that the client connected to */
  g_free (client->server_ip);
  client->server_ip = g_strndup (ip, sizeof (ip));

  GST_INFO ("client %p connected to server ip %s, ipv6 = %d", client,
      client->server_ip, client->is_ipv6);

  url = gst_rtsp_connection_get_url (conn);
  GST_INFO ("added new client %p ip %s:%d", client, url->host, url->port);

  client->wfdsink_ip = url->host;
  GST_INFO ("WFD Sink IP : %s", client->wfdsink_ip);

  client->connection = conn;
  /* create watch for the connection and attach */
  client->watch = gst_rtsp_watch_new (client->connection, &watch_funcs, client, (GDestroyNotify) client_watch_notify);

#if 0
  /* find the context to add the watch */
  if ((source = g_main_current_source ()))
    context = g_source_get_context (source);
  else
    context = NULL;
 #else
     context = g_source_get_context (source);

#endif

  GST_DEBUG (" source = %p", source);


  GST_INFO ("attaching to context %p", context);


  client->watchid = gst_rtsp_watch_attach (client->watch, context);
  gst_rtsp_watch_unref (client->watch);

  return TRUE;

  /* ERRORS */
accept_failed:
  {
    gchar *str = gst_rtsp_strresult (res);

    GST_ERROR ("Could not accept client on server socket %d: %s", sock, str);
    g_free (str);
    return FALSE;
  }
getpeername_failed:
  {
    GST_ERROR ("getpeername failed: %s", g_strerror (errno));
    return FALSE;
  }
getnameinfo_failed:
  {
    GST_ERROR ("getnameinfo failed: %s", g_strerror (errno));
    return FALSE;
  }
}
#endif

/**
* prepare_request:
* @client: client object
* @request : requst message to be prepared
* @method : RTSP method of the request
* @url : url need to be in the request
* @message_type : WFD message type
* @trigger_type : trigger method to be used for M5 mainly
*
* Prepares request based on @method & @message_type
*
* Returns: a #GstRTSPResult.
*/
static GstRTSPResult
prepare_request (GstRTSPClient *client, GstRTSPMessage *request, 
  GstRTSPMethod method, gchar *url, WFDMessageType message_type, WFDTrigger trigger_type)
{
  GstRTSPResult res = GST_RTSP_OK;
  gchar *str = NULL;
  WFDResult wfd_res = WFD_OK;
  GString *cseqstr;
  /* initialize the request */
  res = gst_rtsp_message_init_request (request, method, url);
  if (res < 0) {
    GST_ERROR ("init request failed");
    return res;
  }

  client->cseq++;
  cseqstr = g_string_new ("");
  g_string_append_printf (cseqstr,"%d",client->cseq);
  GST_DEBUG ("CSeq value has been incremented to %d", client->cseq);
  res = gst_rtsp_message_add_header (request, GST_RTSP_HDR_CSEQ, g_string_free (cseqstr, FALSE));
  if (res < 0) {
    GST_ERROR ("Failed to add header");
    return res;
  }

  switch (method) {

    /* Prepare OPTIONS request to send */
    case GST_RTSP_OPTIONS: {
      /* add wfd specific require filed "org.wfa.wfd1.0" */
      str = g_strdup ("org.wfa.wfd1.0");
      res = gst_rtsp_message_add_header (request, GST_RTSP_HDR_REQUIRE, str);
      if (res < 0) {
        GST_ERROR ("Failed to add header");
        return res;
      }
      g_free (str);
      str = NULL;
      break;
    }

    /* Prepare GET_PARAMETER request */
    case GST_RTSP_GET_PARAMETER: {
      gchar *msg;
      guint msglen = 0;
      GString *msglength;
      WFDMessage *msg3;

      /* add content type */
      res = gst_rtsp_message_add_header (request, GST_RTSP_HDR_CONTENT_TYPE, "text/parameters");
      if (res < 0) {
        GST_ERROR ("Failed to add header");
        return res;
      }

      /* create M3 message to be sent in the request */
      wfd_res = wfdconfig_message_new(&msg3);
      if (wfd_res != WFD_OK) {
        GST_ERROR_OBJECT (client, "Failed to create wfd message...");
        res = GST_RTSP_ERROR;
        goto error;
      }

      wfd_res = wfdconfig_message_init(msg3);
      if (wfd_res != WFD_OK) {
        GST_ERROR_OBJECT (client, "Failed to init wfd message...");
        res = GST_RTSP_ERROR;
        goto error;
      }

      /* set the supported audio formats by the WFD server*/
      wfd_res = wfdconfig_set_supported_audio_format(msg3, WFD_AUDIO_UNKNOWN, WFD_FREQ_UNKNOWN,WFD_CHANNEL_UNKNOWN, 0, 0);
      if (wfd_res != WFD_OK) {
        GST_ERROR_OBJECT (client, "Failed to set supported audio formats on wfd message...");
        res = GST_RTSP_ERROR;
        goto error;
      }

      /* set the supported Video formats by the WFD server*/
      wfd_res = wfdconfig_set_supported_video_format(msg3, WFD_VIDEO_UNKNOWN, WFD_VIDEO_CEA_RESOLUTION, WFD_CEA_UNKNOWN,
                        WFD_CEA_UNKNOWN, WFD_VESA_UNKNOWN,WFD_HH_UNKNOWN, WFD_H264_UNKNOWN_PROFILE,
                        WFD_H264_LEVEL_UNKNOWN,0,0, 0, 0, 0, 0);
      if (wfd_res != WFD_OK) {
        GST_ERROR_OBJECT (client, "Failed to set supported video formats on wfd message...");
        res = GST_RTSP_ERROR;
        goto error;
      }

      GST_DEBUG ("wfdconfig_set_display_EDID...");
      wfd_res = wfdconfig_set_display_EDID(msg3, 0, 0, NULL);
      if (wfd_res != WFD_OK) {
        GST_ERROR_OBJECT (client, "Failed to set display EDID type on wfd message...");
        res = GST_RTSP_ERROR;
        goto error;
      }
      GST_DEBUG ("wfdconfig_set_contentprotection_type...");
      wfd_res = wfdconfig_set_contentprotection_type(msg3, WFD_HDCP_NONE, 0);
      if (wfd_res != WFD_OK) {
        GST_ERROR_OBJECT (client, "Failed to set supported content protection type on wfd message...");
        res = GST_RTSP_ERROR;
        goto error;
      }
      /*GST_DEBUG ("wfdconfig_set_coupled_sink...");
      wfd_res = wfdconfig_set_coupled_sink(msg3, WFD_SINK_UNKNOWN, NULL);
      if (wfd_res != WFD_OK) {
        GST_ERROR_OBJECT (client, "Failed to set coupled sink type on wfd message...");
        res = GST_RTSP_ERROR;
        goto error;
      }
      GST_DEBUG ("wfdconfig_set_I2C_port...");
      wfd_res = wfdconfig_set_I2C_port(msg3, FALSE, 0);
      if (wfd_res != WFD_OK) {
        GST_ERROR_OBJECT (client, "Failed to set coupled sink type on wfd message...");
        res = GST_RTSP_ERROR;
        goto error;
      }
      GST_DEBUG ("wfdconfig_set_uibc_capability...");
      wfd_res = wfdconfig_set_uibc_capability(msg3, WFD_UIBC_INPUT_CAT_UNKNOWN, WFD_UIBC_INPUT_TYPE_UNKNOWN, NULL, 0, 0);
      if (wfd_res != WFD_OK) {
        GST_ERROR_OBJECT (client, "Failed to set coupled sink type on wfd message...");
        res = GST_RTSP_ERROR;
        goto error;
      }
      GST_DEBUG ("wfdconfig_set_connector_type...");
      wfd_res = wfdconfig_set_connector_type(msg3, WFD_CONNECTOR_NO);
      if (wfd_res != WFD_OK) {
        GST_ERROR_OBJECT (client, "Failed to set coupled sink type on wfd message...");
        res = GST_RTSP_ERROR;
        goto error;
      } */
#ifdef STANDBY_RESUME_CAPABILITY
      GST_DEBUG ("wfdconfig_set_standby_resume_capability...");
      wfd_res = wfdconfig_set_standby_resume_capability(msg3, FALSE);
      if (wfd_res != WFD_OK) {
        GST_ERROR_OBJECT (client, "Failed to set coupled sink type on wfd message...");
        res = GST_RTSP_ERROR;
        goto error;
      }
#endif
      /* set the preffered RTP ports for the WFD server*/
      wfd_res = wfdconfig_set_prefered_RTP_ports(msg3, WFD_RTSP_TRANS_UNKNOWN, WFD_RTSP_PROFILE_UNKNOWN,
                        WFD_RTSP_LOWER_TRANS_UNKNOWN, 0, 0);
      if (wfd_res != WFD_OK) {
        GST_ERROR_OBJECT (client, "Failed to set supported video formats on wfd message...");
        res = GST_RTSP_ERROR;
        goto error;
      }
      GST_DEBUG ("wfdconfig_parameter_names_as_text...");

/*
      wfd_res = wfdconfig_message_dump(msg3);
      if (wfd_res != WFD_OK) {
        GST_ERROR_OBJECT (client, "Failed to set supported video formats on wfd message...");
        res = GST_RTSP_ERROR;
        goto error;
      }*/

      msg = wfdconfig_parameter_names_as_text(msg3);
      if (msg == NULL) {
        GST_ERROR_OBJECT (client, "Failed to get wfd message as text...");
        res = GST_RTSP_ERROR;
        goto error;
      }
      GST_DEBUG ("wfdconfig_message_as_text...");
      msglen = strlen(msg);
      msglength = g_string_new ("");
      g_string_append_printf (msglength,"%d",msglen);
      GST_DEBUG("M3 server side message body: %s", msg);

      /* add content-length type */
      res = gst_rtsp_message_add_header (request, GST_RTSP_HDR_CONTENT_LENGTH, g_string_free (msglength, FALSE));
      if (res != GST_RTSP_OK) {
        GST_ERROR_OBJECT (client, "Failed to add header to rtsp message...");
        goto error;
      }

      /* adding wfdconfig data to request */
      res = gst_rtsp_message_set_body (request, (guint8*)msg, msglen);
      if (res != GST_RTSP_OK) {
        GST_ERROR_OBJECT (client, "Failed to add header to rtsp message...");
        goto error;
      }

      wfdconfig_message_free(msg3);
      break;
    }

    /* Prepare SET_PARAMETER request */
    case GST_RTSP_SET_PARAMETER: {
      /* add content type */
      res = gst_rtsp_message_add_header (request, GST_RTSP_HDR_CONTENT_TYPE, "text/parameters");
      if (res != GST_RTSP_OK) {
        GST_ERROR_OBJECT (client, "Failed to add header to rtsp request...");
        goto error;
      }

      switch (message_type) {
        case WFD_MESSAGE_4: {
          gchar *msg;
          guint msglen = 0;
          GString *msglength;
          WFDMessage *msg4;

          GString *lines;

          /* create M4 message to be sent in the request */
          wfd_res = wfdconfig_message_new(&msg4);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to create wfd message...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          wfd_res = wfdconfig_message_init(msg4);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to init wfd message...");
            res = GST_RTSP_ERROR;
            goto error;
          }
      lines = g_string_new ("");
      g_string_append_printf (lines,"rtsp://");
      g_string_append_printf (lines, client->server_ip);
      g_string_append_printf (lines,"/wfd1.0/streamid=0");
      wfd_res = wfdconfig_set_presentation_url(msg4, g_string_free (lines, FALSE), NULL);
      if (wfd_res != WFD_OK) {
      GST_ERROR_OBJECT (client, "Failed to set preffered video formats...");
      res = GST_RTSP_ERROR;
      goto error;
      }

          /* set the preffered audio formats for the WFD server*/
          {
            WFDAudioFormats taudiocodec = WFD_AUDIO_UNKNOWN;
        WFDAudioFreq taudiofreq = WFD_FREQ_UNKNOWN;
        WFDAudioChannels taudiochannels = WFD_CHANNEL_UNKNOWN;
        if(client->caCodec & WFD_AUDIO_AC3) taudiocodec = WFD_AUDIO_AAC;  // TODO Currently AC3 encoder is not present
        else if(client->caCodec & WFD_AUDIO_AAC) taudiocodec = WFD_AUDIO_AAC;
        else if(client->caCodec & WFD_AUDIO_LPCM) taudiocodec = WFD_AUDIO_LPCM;
        client->caCodec = taudiocodec;

      if(client->cFreq & WFD_FREQ_48000) taudiofreq = WFD_FREQ_48000;
        else if(client->cFreq & WFD_FREQ_44100) taudiofreq = WFD_FREQ_44100;
        client->cFreq = taudiofreq;

        if(client->cChanels & WFD_CHANNEL_8) taudiochannels = WFD_CHANNEL_2; // TODO Currently only 2 channels is present
        else if(client->cChanels & WFD_CHANNEL_6) taudiochannels = WFD_CHANNEL_2;
        else if(client->cChanels & WFD_CHANNEL_4) taudiochannels = WFD_CHANNEL_2;
        else if(client->cChanels & WFD_CHANNEL_2) taudiochannels = WFD_CHANNEL_2;
            client->cChanels = taudiochannels;

            wfd_res = wfdconfig_set_prefered_audio_format(msg4, taudiocodec, taudiofreq, taudiochannels, client->cBitwidth, client->caLatency);
            if (wfd_res != WFD_OK) {
              GST_ERROR_OBJECT (client, "Failed to set preffered audio formats...");
              res = GST_RTSP_ERROR;
              goto error;
            }
          }
      {
      WFDVideoCEAResolution tcCEAResolution = WFD_CEA_UNKNOWN;
      WFDVideoVESAResolution tcVESAResolution = WFD_VESA_UNKNOWN;
      WFDVideoHHResolution tcHHResolution = WFD_HH_UNKNOWN;
      WFDVideoH264Profile tcProfile;
      WFDVideoH264Level tcLevel;
            /* set the preffered video formats for the WFD server*/
        tcCEAResolution = WFD_CEA_1280x720P30; // TODO need to fetch from INI file
        client->cvCodec = WFD_VIDEO_H264;
        client->cProfile = tcProfile = WFD_H264_BASE_PROFILE; // TODO need to fetch from INI file
        client->cLevel = tcLevel = WFD_H264_LEVEL_3_1; // TODO need to fetch from INI file
        client->cMaxWidth = 1280;
      client->cMaxHeight = 720;
            wfd_res = wfdconfig_set_prefered_video_format(msg4, client->cvCodec, WFD_VIDEO_CEA_RESOLUTION, WFD_CEA_UNKNOWN,
              tcCEAResolution, tcVESAResolution, tcHHResolution, tcProfile,
                            tcLevel, client->cvLatency, client->cMaxHeight,client->cMaxWidth, client->cmin_slice_size, client->cslice_enc_params, client->cframe_rate_control);
            if (wfd_res != WFD_OK) {
              GST_ERROR_OBJECT (client, "Failed to set preffered video formats...");
              res = GST_RTSP_ERROR;
              goto error;
            }
          }
          GST_DEBUG("wfd config set presentation URL %s",url);

          /* set the preffered RTP ports for the WFD server*/
          GST_LOG("Port are %d, %d\n", client->crtp_port0, client->crtp_port1);
          wfd_res = wfdconfig_set_prefered_RTP_ports(msg4, WFD_RTSP_TRANS_RTP, WFD_RTSP_PROFILE_AVP,
              WFD_RTSP_LOWER_TRANS_UDP, client->crtp_port0, client->crtp_port1);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to set preffered RTP ports...");
            res = GST_RTSP_ERROR;
            goto error;
          }
	   /*set the preffered hdcp version and port for the WFD server*/
	   if (client->hdcp_support) {
            GST_DEBUG("hdcp version =%d, tcp port = %d", client->hdcp_version, client->hdcp_tcpport);
            wfd_res = wfdconfig_set_contentprotection_type(msg4, client->hdcp_version, client->hdcp_tcpport);
            if (wfd_res != WFD_OK) {
              GST_ERROR_OBJECT (client, "Failed to set supported content protection type on wfd message...");
              res = GST_RTSP_ERROR;
              goto error;
            }
	   }
#ifdef STANDBY_RESUME_CAPABILITY	   
       if(client->standby_resume_capability_support){
	     wfd_res = wfdconfig_set_standby_resume_capability(msg4, TRUE);
	     if (wfd_res != WFD_OK) {
              GST_ERROR_OBJECT (client, "Failed to set supported standby resume capability  type on wfd message...");
              res = GST_RTSP_ERROR;
              goto error;
            }
	   }
#endif
          wfd_res = wfdconfig_message_dump(msg4);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to dump wfd message...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          msg = wfdconfig_message_as_text(msg4);
          if (msg == NULL) {
            GST_ERROR_OBJECT (client, "Failed to get wfd message as text...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          msglen = strlen(msg);
          msglength = g_string_new ("");
          g_string_append_printf (msglength,"%d",msglen);
          GST_DEBUG("M4 message body server side : %s", msg);

          /* add content-length type */
          res = gst_rtsp_message_add_header (request, GST_RTSP_HDR_CONTENT_LENGTH, g_string_free (msglength, FALSE));
          if (res != GST_RTSP_OK) {
            GST_ERROR_OBJECT (client, "Failed to add header to rtsp request...");
            goto error;
          }

          /* adding wfdconfig data to request */
          res = gst_rtsp_message_set_body (request, (guint8*)msg, msglen);
          if (res != GST_RTSP_OK) {
            GST_ERROR_OBJECT (client, "Failed to set body to rtsp request...");
            goto error;
          }

          wfdconfig_message_free(msg4);
        }
        break;
        case WFD_MESSAGE_5: {
          WFDMessage *msg5;
          gchar *msg;
          guint msglen = 0;
          GString *msglength;

          /* create M4 message to be sent in the request */
          wfd_res = wfdconfig_message_new(&msg5);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to create wfd message...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          wfd_res = wfdconfig_message_init(msg5);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to init wfd message...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          /* preparing SETUP trigger message. After client receving this request, client has to send SETUP request */
          wfd_res = wfdconfig_set_trigger_type(msg5, trigger_type);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to trigger type...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          wfd_res = wfdconfig_message_dump(msg5);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to dump wfd message...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          msg = wfdconfig_message_as_text(msg5);
          if (msg == NULL) {
            GST_ERROR_OBJECT (client, "Failed to get wfd message as text...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          msglen = strlen(msg);
          msglength = g_string_new ("");
          g_string_append_printf (msglength,"%d",msglen);

          GST_DEBUG ("M5 trigger message body: %s", msg);

          /* add content-length type */
          res = gst_rtsp_message_add_header (request, GST_RTSP_HDR_CONTENT_LENGTH, g_string_free (msglength, FALSE));
          if (res != GST_RTSP_OK) {
            GST_ERROR_OBJECT (client, "Failed to add header to rtsp request...");
            goto error;
          }

          /* adding wfdconfig data to request */
          res = gst_rtsp_message_set_body (request, (guint8*)msg, msglen);
          if (res != GST_RTSP_OK) {
            GST_ERROR_OBJECT (client, "Failed to set body to rtsp request...");
            goto error;
          }

          wfdconfig_message_free(msg5);
          break;
        }
        case WFD_MESSAGE_12: {
          WFDMessage *msg12;
          gchar *msg;
          guint msglen = 0;
          GString *msglength;

          /* create M4 message to be sent in the request */
          wfd_res = wfdconfig_message_new(&msg12);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to create wfd message...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          wfd_res = wfdconfig_message_init(msg12);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to init wfd message...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          /* preparing SETUP trigger message. After client receving this request, client has to send SETUP request */
          wfd_res = wfdconfig_set_standby(msg12, TRUE);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to trigger type...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          wfd_res = wfdconfig_message_dump(msg12);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to dump wfd message...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          msg = wfdconfig_message_as_text(msg12);
          if (msg == NULL) {
            GST_ERROR_OBJECT (client, "Failed to get wfd message as text...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          msglen = strlen(msg);
          msglength = g_string_new ("");
          g_string_append_printf (msglength,"%d",msglen);

          GST_DEBUG ("M12 standby message body: %s", msg);

          /* add content-length type */
          res = gst_rtsp_message_add_header (request, GST_RTSP_HDR_CONTENT_LENGTH, g_string_free (msglength, FALSE));
          if (res != GST_RTSP_OK) {
            GST_ERROR_OBJECT (client, "Failed to add header to rtsp request...");
            goto error;
          }

          /* adding wfdconfig data to request */
          res = gst_rtsp_message_set_body (request, (guint8*)msg, msglen);
          if (res != GST_RTSP_OK) {
            GST_ERROR_OBJECT (client, "Failed to set body to rtsp request...");
            goto error;
          }

          wfdconfig_message_free(msg12);
          break;
        }
        case WFD_MESSAGE_14: {
          WFDMessage *msg14;
          gchar *msg;
          guint msglen = 0;
          GString *msglength;
          WFDHIDCTypePathPair inp_pair[] = {{WFD_UIBC_INPUT_TYPE_KEYBOARD,WFD_UIBC_INPUT_PATH_USB},{WFD_UIBC_INPUT_TYPE_MOUSE,WFD_UIBC_INPUT_PATH_BT},{WFD_UIBC_INPUT_TYPE_MULTITOUCH,WFD_UIBC_INPUT_PATH_WIFI},{WFD_UIBC_INPUT_TYPE_GESTURE,WFD_UIBC_INPUT_PATH_BT}};

          /* create M4 message to be sent in the request */
          wfd_res = wfdconfig_message_new(&msg14);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to create wfd message...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          wfd_res = wfdconfig_message_init(msg14);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to init wfd message...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          /* preparing SETUP trigger message. After client receving this request, client has to send SETUP request */
          wfd_res = wfdconfig_set_uibc_capability(msg14, WFD_UIBC_INPUT_CAT_GENERIC,
               WFD_UIBC_INPUT_TYPE_MOUSE|WFD_UIBC_INPUT_TYPE_MULTITOUCH|WFD_UIBC_INPUT_TYPE_GESTURE, inp_pair, 4, 8558);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to trigger type...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          wfd_res = wfdconfig_message_dump(msg14);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to dump wfd message...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          msg = wfdconfig_message_as_text(msg14);
          if (msg == NULL) {
            GST_ERROR_OBJECT (client, "Failed to get wfd message as text...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          msglen = strlen(msg);
          msglength = g_string_new ("");
          g_string_append_printf (msglength,"%d",msglen);

          GST_DEBUG ("M14 UIBC message body: %s", msg);

          /* add content-length type */
          res = gst_rtsp_message_add_header (request, GST_RTSP_HDR_CONTENT_LENGTH, g_string_free (msglength, FALSE));
          if (res != GST_RTSP_OK) {
            GST_ERROR_OBJECT (client, "Failed to add header to rtsp request...");
            goto error;
          }

          /* adding wfdconfig data to request */
          res = gst_rtsp_message_set_body (request, (guint8*)msg, msglen);
          if (res != GST_RTSP_OK) {
            GST_ERROR_OBJECT (client, "Failed to set body to rtsp request...");
            goto error;
          }

          wfdconfig_message_free(msg14);
          break;
        }
        case WFD_MESSAGE_15: {
          WFDMessage *msg15;
          gchar *msg;
          guint msglen = 0;
          GString *msglength;

          /* create M4 message to be sent in the request */
          wfd_res = wfdconfig_message_new(&msg15);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to create wfd message...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          wfd_res = wfdconfig_message_init(msg15);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to init wfd message...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          /* preparing SETUP trigger message. After client receving this request, client has to send SETUP request */
          wfd_res = wfdconfig_set_uibc_status(msg15, TRUE);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to trigger type...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          wfd_res = wfdconfig_message_dump(msg15);
          if (wfd_res != WFD_OK) {
            GST_ERROR_OBJECT (client, "Failed to dump wfd message...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          msg = wfdconfig_message_as_text(msg15);
          if (msg == NULL) {
            GST_ERROR_OBJECT (client, "Failed to get wfd message as text...");
            res = GST_RTSP_ERROR;
            goto error;
          }

          msglen = strlen(msg);
          msglength = g_string_new ("");
          g_string_append_printf (msglength,"%d",msglen);

          GST_DEBUG ("M15 UIBC enable message body: %s", msg);

          /* add content-length type */
          res = gst_rtsp_message_add_header (request, GST_RTSP_HDR_CONTENT_LENGTH, g_string_free (msglength, FALSE));
          if (res != GST_RTSP_OK) {
            GST_ERROR_OBJECT (client, "Failed to add header to rtsp request...");
            goto error;
          }

          /* adding wfdconfig data to request */
          res = gst_rtsp_message_set_body (request, (guint8*)msg, msglen);
          if (res != GST_RTSP_OK) {
            GST_ERROR_OBJECT (client, "Failed to set body to rtsp request...");
            goto error;
          }

          wfdconfig_message_free(msg15);
          break;
        }
        default:
          GST_ERROR_OBJECT (client, "Unhandled WFD message type...");
          return GST_RTSP_EINVAL;
          break;
      }
    }
    break;

    default:
      GST_ERROR_OBJECT (client, "Unhandled method...");
      return GST_RTSP_EINVAL;
      break;
  }

  return res;

error:
  return res;
}

/**
* prepare_response:
* @client: client object
* @request : requst message received
* @response : response to be prepare based on request
* @method : RTSP method
*
* prepare response to the request based on @method & @message_type
*
* Returns: a #GstRTSPResult.
*/
static GstRTSPResult
prepare_response (GstRTSPClient *client, GstRTSPMessage *request, GstRTSPMessage *response, GstRTSPMethod method)
{
  GstRTSPResult res = GST_RTSP_OK;

  switch (method) {
    /* prepare OPTIONS response */
    case GST_RTSP_OPTIONS: {
      GstRTSPMethod options;
      gchar *tmp = NULL;
      gchar *str = NULL;
      gchar *user_agent = NULL;

      options = GST_RTSP_OPTIONS |
          GST_RTSP_PAUSE |
          GST_RTSP_PLAY |
          GST_RTSP_SETUP |
          GST_RTSP_GET_PARAMETER | GST_RTSP_SET_PARAMETER | GST_RTSP_TEARDOWN;

      str = gst_rtsp_options_as_text (options);

      /*append WFD specific method */
      tmp = g_strdup (", org.wfa.wfd1.0");
      g_strlcat (str, tmp, strlen (tmp) + strlen (str) + 1);

      gst_rtsp_message_init_response (response, GST_RTSP_STS_OK,
          gst_rtsp_status_as_text (GST_RTSP_STS_OK), request);

      gst_rtsp_message_add_header (response, GST_RTSP_HDR_PUBLIC, str);
      g_free (str);
      str = NULL;
      res = gst_rtsp_message_get_header (request, GST_RTSP_HDR_USER_AGENT, &user_agent, 0);
      if (res == GST_RTSP_OK)
      {
        gst_rtsp_message_add_header (response, GST_RTSP_HDR_USER_AGENT, user_agent);
      }
      else res = GST_RTSP_OK;
      break;
    }
    default:
      GST_ERROR_OBJECT (client, "Unhandled method...");
      return GST_RTSP_EINVAL;
      break;
  }

  return res;
}

/**
* handle_M1_message:
* @client: client object
*
* Handles M1 WFD message.
* This API will send OPTIONS request to WFDSink & waits for the relavent response.
* After getting the response, this will check whether all mandatory messages are supported by the WFDSink or NOT.
*
* Returns: a #GstRTSPResult.
*/

static GstRTSPResult
handle_M1_message (GstRTSPClient * client)
{
  GstRTSPResult res = GST_RTSP_OK;
  GstRTSPMessage request = { 0 };
  GstRTSPMessage response = { 0 };
  gboolean bret = FALSE;

  res = prepare_request (client, &request, GST_RTSP_OPTIONS, "*", WFD_MESSAGE_1, WFD_TRIGGER_UNKNOWN);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to prepare M1 request....\n");
    return res;
  }

  GST_DEBUG_OBJECT (client, "Sending OPTIONS request message (M1)...");

  // TODO: need to add session i.e. 2nd variable
  send_request (client, NULL, &request);

  /* Wait for OPTIONS response (M1 response) */
  res = gst_rtsp_connection_receive (client->connection, &response, NULL);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to receive M1 response....\n");
    return res;
  }

  if (gst_debug_category_get_threshold (rtsp_client_debug) >= GST_LEVEL_LOG) {
    gst_rtsp_message_dump (&response);
  }

  bret = gst_rtsp_client_parse_methods (client, &response);
  if (FALSE == bret) {
    return GST_RTSP_ERROR;
  }

  return res;

}

/**
* handle_M2_message:
* @client: client object
*
* Handles M2 WFD message.
* This API will waits for OPTIONS request from WFDSink & responds to the same.
*
* Returns: a #GstRTSPResult.
*/
static GstRTSPResult
handle_M2_message (GstRTSPClient * client)
{
  GstRTSPResult res = GST_RTSP_OK;
  GstRTSPMessage request = { 0 };
  GstRTSPMessage response = { 0 };
  GstRTSPMethod method;
  const gchar *uristr;
  GstRTSPVersion version;
  GstRTSPClientState state = { NULL };

  /* Wait for OPTIONS request from client (M2 request) */
  res = gst_rtsp_connection_receive (client->connection, &request, NULL);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to receiv M2 request....\n");
    return res;
  }
  /*dump M2 request message*/
  if (gst_debug_category_get_threshold (rtsp_client_debug) >= GST_LEVEL_LOG) {
    gst_rtsp_message_dump (&request);
  }

  /* Parse the request received */
  res = gst_rtsp_message_parse_request (&request, &method, &uristr, &version);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to parse request....\n");
    return res;
  }

  state.request = &request;
  state.response = &response;

  if (version != GST_RTSP_VERSION_1_0) {
    /* we can only handle 1.0 requests */
    send_generic_response (client, GST_RTSP_STS_RTSP_VERSION_NOT_SUPPORTED,
        &state);
    return res;
  }

  if (method != GST_RTSP_OPTIONS) {
    GST_ERROR_OBJECT (client, "Received WRONG request, when server is Waiting for OPTIONS request...");
    return GST_RTSP_ERROR;
  }
#if 0
  /* we always try to parse the url first */
  if (gst_rtsp_url_parse (uristr, &uri) != GST_RTSP_OK) {
    send_generic_response (client, GST_RTSP_STS_BAD_REQUEST, &state);
    return GST_RTSP_ERROR;
  }

  /* not using the uri at this moment */
  gst_rtsp_url_free (uri);
#endif

  GST_DEBUG_OBJECT (client, "Received OPTIONS Request (M2 request)...");

  /* prepare the response for WFDsink request */
  res = prepare_response (client, &request, &response, GST_RTSP_OPTIONS);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to prepare M2 request....\n");
    return res;
  }

  send_response (client, NULL, &response);

  GST_DEBUG_OBJECT (client, "Sent OPTIONS Response (M2 response)...");

  return res;

}

/**
* handle_M3_message:
* @client: client object
*
* Handles M3 WFD message.
* This API will send M3 message (GET_PARAMETER) to WFDSink to query supported formats by the WFDSink.
* After getting supported formats info, this API will set those values on WFDConfigMessage obj
*
* Returns: a #GstRTSPResult.
*/
static GstRTSPResult
handle_M3_message (GstRTSPClient * client)
{
  GstRTSPResult res = GST_RTSP_OK;
  GstRTSPMessage request = { 0 };
  GstRTSPMessage response = { 0 };
  GstRTSPUrl *url = NULL;
  gchar *url_str = NULL;

  client->caCodec = WFD_AUDIO_UNKNOWN;
  client->cFreq = WFD_FREQ_UNKNOWN;
  client->cChanels = WFD_CHANNEL_UNKNOWN;
  client->cBitwidth = 0;
  client->caLatency = 0;
  client->cvCodec = WFD_VIDEO_H264;
  client->cNative = WFD_VIDEO_CEA_RESOLUTION;
  client->cNativeResolution = 0;
  client->cCEAResolution = WFD_CEA_UNKNOWN;
  client->cVESAResolution = WFD_VESA_UNKNOWN;
  client->cHHResolution = WFD_HH_UNKNOWN;
  client->cProfile = WFD_H264_UNKNOWN_PROFILE;
  client->cLevel = WFD_H264_LEVEL_UNKNOWN;
  client->cMaxHeight = 0;
  client->cMaxWidth = 0;
  client->cmin_slice_size = 0;
  client->cslice_enc_params = 0;
  client->cframe_rate_control = 0;
  client->cvLatency = 0;
  client->ctrans = WFD_RTSP_TRANS_UNKNOWN;
  client->cprofile = WFD_RTSP_PROFILE_UNKNOWN;
  client->clowertrans = WFD_RTSP_LOWER_TRANS_UNKNOWN;
  client->crtp_port0 = 0;
  client->crtp_port1 = 0;
  client->hdcp_version = WFD_HDCP_NONE;
  client->hdcp_tcpport = 0;
  client->hdcp_support = FALSE;
#ifdef STANDBY_RESUME_CAPABILITY
  client->standby_resume_capability_support = FALSE;
#endif

  url = gst_rtsp_connection_get_url (client->connection);
  if (url == NULL) {
    GST_ERROR_OBJECT (client, "Failed to get connection URL");
    return GST_RTSP_ERROR;
  }

  url_str = gst_rtsp_url_get_request_uri (url);
  if (url_str == NULL) {
    GST_ERROR_OBJECT (client, "Failed to get connection URL");
    return GST_RTSP_ERROR;
  }

  res = prepare_request (client, &request, GST_RTSP_GET_PARAMETER, url_str, WFD_MESSAGE_3, WFD_TRIGGER_UNKNOWN);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to prepare M3 request....\n");
    return res;
  }

  GST_DEBUG_OBJECT (client, "Sending GET_PARAMETER request message (M3)...");

  // TODO: need to add session i.e. 2nd variable
  send_request (client, NULL, &request);

  /* Wait for the response */
  res = gst_rtsp_connection_receive (client->connection, &response, NULL);
  if (GST_RTSP_OK != res) {
    GST_ERROR ("Failed to received response....\n");
    return FALSE;
  }

  if (gst_debug_category_get_threshold (rtsp_client_debug) >= GST_LEVEL_LOG) {
    gst_rtsp_message_dump (&response);
  }

  /* parsing the GET_PARAMTER response */
  {
    gchar *data = NULL;
    guint size=0;
    WFDMessage *msg3res;
    WFDResult wfd_res = WFD_OK;

    res = gst_rtsp_message_get_body (&response, (guint8**)&data, &size);
    if (res != GST_RTSP_OK) {
      GST_ERROR_OBJECT (client, "Failed to get body of response...");
      goto error;
    }

    /* create M3 response message */
    wfd_res = wfdconfig_message_new(&msg3res);
    if (wfd_res != WFD_OK) {
      GST_ERROR_OBJECT (client, "Failed to prepare wfd message...");
      goto error;
    }

    wfd_res = wfdconfig_message_init(msg3res);
    if (wfd_res != WFD_OK) {
      GST_ERROR_OBJECT (client, "Failed to init wfd message...");
      goto error;
    }

    wfd_res = wfdconfig_message_parse_buffer((guint8*)data,size,msg3res);
    if (wfd_res != WFD_OK) {
      GST_ERROR_OBJECT (client, "Failed to init wfd message...");
      goto error;
    }

    GST_DEBUG_OBJECT(client, "M3 response server side message body: %s", wfdconfig_message_as_text(msg3res));

    /* Get the audio formats supported by WFDSink */
    if (msg3res->audio_codecs) {
      wfd_res = wfdconfig_get_supported_audio_format(msg3res, &client->caCodec, &client->cFreq, &client->cChanels, &client->cBitwidth, &client->caLatency);
      if (wfd_res != WFD_OK) {
        GST_WARNING_OBJECT (client, "Failed to get wfd support audio formats...");
        goto error;
      }
    }


    /* Get the Video formats supported by WFDSink */
    wfd_res = wfdconfig_get_supported_video_format(msg3res, &client->cvCodec, &client->cNative, &client->cNativeResolution,
                      (guint64*)&client->cCEAResolution, (guint64*)&client->cVESAResolution, (guint64*)&client->cHHResolution,
                      &client->cProfile, &client->cLevel, &client->cvLatency, &client->cMaxHeight,
                      &client->cMaxWidth, &client->cmin_slice_size, &client->cslice_enc_params, &client->cframe_rate_control);
    if (wfd_res != WFD_OK) {
      GST_WARNING_OBJECT (client, "Failed to get wfd supported video formats...");
      goto error;
    }

    if (msg3res->client_rtp_ports) {
      /* Get the RTP ports preferred by WFDSink */
      wfd_res = wfdconfig_get_prefered_RTP_ports(msg3res, &client->ctrans, &client->cprofile, &client->clowertrans, &client->crtp_port0, &client->crtp_port1);
      if (wfd_res != WFD_OK) {
        GST_WARNING_OBJECT (client, "Failed to get wfd prefered RTP ports...");
        goto error;
      }
    }

    if (msg3res->display_edid) {
      gboolean edid_supported = FALSE;
      guint32 edid_block_count = 0;
      gchar *edid_payload = NULL;
      /* Get the display edid preferred by WFDSink */
      wfd_res = wfdconfig_get_display_EDID(msg3res, &edid_supported, &edid_block_count, &edid_payload);
      if (wfd_res != WFD_OK) {
        GST_WARNING_OBJECT (client, "Failed to get wfd display edid...");
        goto error;
      }
      GST_DEBUG_OBJECT (client, " edid supported: %d edid_block_count: %d", edid_supported, edid_block_count);
      GST_DEBUG_OBJECT (client, "set_edid_info");
      set_edid_info(edid_payload, TRUE);
    }
   if (msg3res->content_protection) {
     /*Get the hdcp version and tcp port by WFDSink*/
     wfd_res = wfdconfig_get_contentprotection_type(msg3res, &client->hdcp_version, &client->hdcp_tcpport);
     GST_DEBUG("hdcp version =%d, tcp port = %d", client->hdcp_version, client->hdcp_tcpport);
     if (client->hdcp_version > 0 && client->hdcp_tcpport > 0)
       client->hdcp_support = TRUE;

      if (wfd_res != WFD_OK) {
        GST_WARNING_OBJECT (client, "Failed to get wfd content protection...");
        goto error;
      }
   }
#ifdef STANDBY_RESUME_CAPABILITY
   if (msg3res->standby_resume_capability) {
     /*Get the standby_resume_capability value by WFDSink*/
     wfd_res = wfdconfig_get_standby_resume_capability(msg3res, &client->standby_resume_capability_support );
     if (wfd_res != WFD_OK) {
       GST_WARNING_OBJECT (client, "Failed to get wfd standby resume capability...");
       goto error;
     }
   }
#endif
    wfdconfig_message_dump(msg3res);
  }

  return res;

error:
  return GST_RTSP_ERROR;

}

/**
* handle_M4_message:
* @client: client object
*
* Handles M4 WFD message.
* This API will send M4 message (SET_PARAMETER) to WFDSink to set the format to be used in session
*
* Returns: a #GstRTSPResult.
*/
static GstRTSPResult
handle_M4_message (GstRTSPClient * client)
{
  GstRTSPResult res = GST_RTSP_OK;
  GstRTSPMessage request = { 0 };
  GstRTSPMessage response = { 0 };
  GstRTSPUrl *url = NULL;
  gchar *url_str = NULL;

  url = gst_rtsp_connection_get_url (client->connection);
  if (url == NULL) {
    GST_ERROR_OBJECT (client, "Failed to get connection URL");
    return GST_RTSP_ERROR;
  }

  url_str = gst_rtsp_url_get_request_uri (url);
  if (url_str == NULL) {
    GST_ERROR_OBJECT (client, "Failed to get connection URL");
    return GST_RTSP_ERROR;
  }

  /* prepare the request for M4 message */
  res = prepare_request (client, &request, GST_RTSP_SET_PARAMETER, url_str, WFD_MESSAGE_4, WFD_TRIGGER_UNKNOWN);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to prepare M4 request....\n");
    return res;
  }

  GST_DEBUG_OBJECT (client, "Sending SET_PARAMETER request message (M4)...");

  // TODO: need to add session i.e. 2nd variable
  send_request (client, NULL, &request);

  /* Wait for the M4 response from WFDSink */
  res = gst_rtsp_connection_receive (client->connection, &response, NULL);
  if (GST_RTSP_OK != res) {
    GST_ERROR ("Failed to received response....\n");
    return FALSE;
  }

  if (gst_debug_category_get_threshold (rtsp_client_debug) >= GST_LEVEL_LOG) {
    gst_rtsp_message_dump (&response);
  }

  return res;

}


/**
* handle_M5_message:
* @client: client object
*
* Handles M5 WFD message.
* This API will send M5 SETUP trigger message using SET_PARAMETER to WFDSink to intimate that client request SETUP now.
*
* Returns: a #GstRTSPResult.
*/
static GstRTSPResult
handle_M5_message (GstRTSPClient * client, WFDTrigger trigger_type)
{
  GstRTSPResult res = GST_RTSP_OK;
  GstRTSPMessage request = { 0 };
  GstRTSPMessage response = { 0 };
  GstRTSPUrl *url = NULL;
  gchar *url_str = NULL;
  GstRTSPSession *session = NULL;

  url = gst_rtsp_connection_get_url (client->connection);
  if (url == NULL) {
    GST_ERROR_OBJECT (client, "Failed to get connection URL");
    return GST_RTSP_ERROR;
  }

  url_str = gst_rtsp_url_get_request_uri (url);
  if (url_str == NULL) {
    GST_ERROR_OBJECT (client, "Failed to get connection URL");
    return GST_RTSP_ERROR;
  }

  /* prepare the request for M5 message */
  res = prepare_request (client, &request, GST_RTSP_SET_PARAMETER, url_str, WFD_MESSAGE_5, trigger_type);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to prepare M5 request....\n");
    return res;
  }

  GST_DEBUG_OBJECT (client, "Sending SET_PARAMETER request message (M5)...");

  // TODO: need to add session i.e. 2nd variable
  if (client->sessionid) {
    session = gst_rtsp_session_pool_find (client->session_pool, client->sessionid);
    GST_INFO_OBJECT (client, "session = %p & sessionid = %s", session, session->sessionid);
  }
  send_request (client, session, &request);

  /* Wait for the M5 response from WFDSink */
  res = gst_rtsp_connection_receive (client->connection, &response, NULL);
  if (GST_RTSP_OK != res) {
    GST_ERROR ("Failed to received response....\n");
    return FALSE;
  }

  if (gst_debug_category_get_threshold (rtsp_client_debug) >= GST_LEVEL_LOG) {
    gst_rtsp_message_dump (&response);
  }

  return res;

}


/**
* handle_M12_message:
* @client: client object
*
* Handles M12 WFD message.
* This API will send M5 SETUP trigger message using SET_PARAMETER to WFDSink to intimate that client request SETUP now.
*
* Returns: a #GstRTSPResult.
*/
static GstRTSPResult
handle_M12_message (GstRTSPClient * client)
{
  GstRTSPResult res = GST_RTSP_OK;
  GstRTSPMessage request = { 0 };
  GstRTSPMessage response = { 0 };
  GstRTSPUrl *url = NULL;
  gchar *url_str = NULL;
  GstRTSPSession *session = NULL;
  GstRTSPSessionMedia *media;
  GstRTSPStatusCode code;

  url = gst_rtsp_connection_get_url (client->connection);
  if (url == NULL) {
    GST_ERROR_OBJECT (client, "Failed to get connection URL");
    return GST_RTSP_ERROR;
  }

  url_str = gst_rtsp_url_get_request_uri (url);
  if (url_str == NULL) {
    GST_ERROR_OBJECT (client, "Failed to get connection URL");
    return GST_RTSP_ERROR;
  }

  /* prepare the request for M12 message */
  res = prepare_request (client, &request, GST_RTSP_SET_PARAMETER, url_str, WFD_MESSAGE_12, WFD_TRIGGER_UNKNOWN);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to prepare M12 request....\n");
    return res;
  }

  GST_DEBUG_OBJECT (client, "Sending SET_PARAMETER request message (M12)...");

  // TODO: need to add session i.e. 2nd variable
  if (client->sessionid) {
    session = gst_rtsp_session_pool_find (client->session_pool, client->sessionid);
    GST_INFO_OBJECT (client, "session = %p & sessionid = %s", session, session->sessionid);
  }
  send_request (client, session, &request);

  /* Wait for the M12 response from WFDSink */
  res = gst_rtsp_connection_receive (client->connection, &response, NULL);
  if (GST_RTSP_OK != res) {
    GST_ERROR ("Failed to received response....\n");
    return FALSE;
  }

  if (gst_debug_category_get_threshold (rtsp_client_debug) >= GST_LEVEL_LOG) {
    gst_rtsp_message_dump (&response);
  }
  if (!(session = gst_rtsp_session_pool_find (client->session_pool, client->sessionID)))
  {
    GST_ERROR_OBJECT (client, "Failed to handle B1 message...");
    return FALSE;
  }
  GST_DEBUG ("sessid = %s", client->sessionID);

  media = session->medias->data;

  /* unlink the all TCP callbacks */
  unlink_session_streams (client, session, media);

  /* then pause sending */
  gst_rtsp_session_media_set_state (media, GST_STATE_PAUSED);

  /* the state is now READY */
  media->state = GST_RTSP_STATE_READY;

  return res;

}
/**
* gst_rtsp_client_negotiate:
* @client: client object
*
* This will handle capability negotiation part of WFD session
*
* Returns: a #gboolean. return TRUE on success / FALSE on failure
*/
gboolean
gst_rtsp_client_negotiate (GstRTSPClient * client)
{
  GstRTSPResult res = GST_RTSP_OK;

  /* handle M1 OPTIONS message */
  res = handle_M1_message (client);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to handle M1 message...");
    goto rtsp_method_failed;
  }

  /* handle M2 OPTIONS message */
  res = handle_M2_message (client);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to handle M2 message...");
    goto rtsp_method_failed;
  }

  /* handle M3 GET_PARAMETER request message */
  res = handle_M3_message (client);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to handle M3 message...");
    goto rtsp_method_failed;
  }

  /* handle M4 SET_PARAMETER request message */
  res = handle_M4_message (client);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to handle M4 message...");
    goto rtsp_method_failed;
  }

  return TRUE;

rtsp_method_failed:
  GST_ERROR_OBJECT (client, "Failed to handle rtsp method : %d", res);
  return FALSE;

}

/**
* gst_rtsp_client_start:
* @client: client object
*
* This will handle M5 message & asks the client to start session
*
* Returns: a #gboolean. return TRUE on success / FALSE on failure
*/
gboolean
gst_rtsp_client_start (GstRTSPClient * client)
{
  GstRTSPResult res = GST_RTSP_OK;

  res = handle_M5_message (client, WFD_TRIGGER_SETUP);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to handle M5 message...");
    return FALSE;
  }

  return TRUE;
}

void
gst_rtsp_client_set_params (GstRTSPClient *client, int videosrc_type, gint session_mode, int videobitrate, gint mtu_size, gchar *infile)
{
  client->videosrc_type = videosrc_type;
  client->session_mode = session_mode;
  client->bitrate = videobitrate;
  client->MTUsize = mtu_size;
  client->infile = g_strdup (infile);

  g_print ("\n\n\nvideo src type = %d & session_mode = %d & filename = %s\n",
    videosrc_type, session_mode, client->infile);
}

gboolean
gst_rtsp_client_pause (GstRTSPClient * client)
{
  GstRTSPResult res = GST_RTSP_OK;

  res = handle_M5_message (client, WFD_TRIGGER_PAUSE);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to handle M5 message...");
    return FALSE;
  }

  return TRUE;
}

gboolean
gst_rtsp_client_resume (GstRTSPClient * client)
{
  GstRTSPResult res = GST_RTSP_OK;

  res = handle_M5_message (client, WFD_TRIGGER_PLAY);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to handle M5 message...");
    return FALSE;
  }

  return TRUE;
}

gboolean
gst_rtsp_client_standby(GstRTSPClient * client)
{
  GstRTSPResult res = GST_RTSP_OK;
  res = handle_M12_message (client);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to handle M12 message...");
    return FALSE;
  }

  return TRUE;
}

gboolean
gst_rtsp_client_stop (GstRTSPClient * client)
{
  GstRTSPResult res = GST_RTSP_OK;

  res = handle_M5_message (client, WFD_TRIGGER_TEARDOWN);
  if (GST_RTSP_OK != res) {
    GST_ERROR_OBJECT (client, "Failed to handle M5 message...");
    return FALSE;
  }

  return TRUE;
}

static gboolean
gst_rtsp_client_parse_methods (GstRTSPClient * client, GstRTSPMessage * response)
{
  GstRTSPHeaderField field;
  gchar *respoptions;
  gchar **options;
  gint indx = 0;
  gint i;
  gboolean found_wfd_method = FALSE;

  /* reset supported methods */
  client->methods = 0;

  /* Try Allow Header first */
  field = GST_RTSP_HDR_ALLOW;
  while (TRUE) {
    respoptions = NULL;
    gst_rtsp_message_get_header (response, field, &respoptions, indx);
    if (indx == 0 && !respoptions) {
      /* if no Allow header was found then try the Public header... */
      field = GST_RTSP_HDR_PUBLIC;
      gst_rtsp_message_get_header (response, field, &respoptions, indx);
    }
    if (!respoptions)
      break;

    /* If we get here, the server gave a list of supported methods, parse
     * them here. The string is like:
     *
     * OPTIONS,  PLAY, SETUP, ...
     */
    options = g_strsplit (respoptions, ",", 0);

    for (i = 0; options[i]; i++) {
      gchar *stripped;
      gint method;

      stripped = g_strstrip (options[i]);
      method = gst_rtsp_find_method (stripped);

      if (!g_ascii_strcasecmp ("org.wfa.wfd1.0", stripped))
        found_wfd_method = TRUE;

      /* keep bitfield of supported methods */
      if (method != GST_RTSP_INVALID)
        client->methods |= method;
    }
    g_strfreev (options);

    indx++;
  }

  if (!found_wfd_method) {
    GST_ERROR_OBJECT (client, "WFD client is not supporting WFD mandatory message : org.wfa.wfd1.0...");
    goto no_required_methods;
  }

  /* Checking mandatory method */
  if (!(client->methods & GST_RTSP_SET_PARAMETER)) {
    GST_ERROR_OBJECT (client, "WFD client is not supporting WFD mandatory message : SET_PARAMETER...");
    goto no_required_methods;
  }

  /* Checking mandatory method */
  if (!(client->methods & GST_RTSP_GET_PARAMETER)) {
    GST_ERROR_OBJECT (client, "WFD client is not supporting WFD mandatory message : GET_PARAMETER...");
    goto no_required_methods;
  }

  if (!(client->methods & GST_RTSP_OPTIONS)) {
    GST_INFO_OBJECT (client, "assuming OPTIONS is supported by client...");
    client->methods |= GST_RTSP_OPTIONS;
  }

  return TRUE;

/* ERRORS */
no_required_methods:
  {
    GST_ELEMENT_ERROR (client, RESOURCE, OPEN_READ, (NULL),
        ("WFD Client does not support mandatory methods."));
  // TODO: temporary hack
    return FALSE;
  }
}


static void
pad_added (GstElement* ele, GstPad* pad, gpointer data)
{
  GstRTSPClient *client = (GstRTSPClient *)data;
  gchar* name = gst_pad_get_name (pad);

  if (name[0] == 'v') {
    GstPad* spad = NULL;

    g_print (" =========== >>>>>>>>>> Received VIDEO pad...\n");

    spad = gst_element_get_pad(GST_ELEMENT(client->srcbin->vparse), "sink");
    if (gst_pad_link(pad, spad)  != GST_PAD_LINK_OK) {
      GST_ERROR ("Failed to link video demuxer pad & video parser sink pad...\n");
      return;
    }
    gst_object_unref(spad);

  } else if (name[0] == 'a') {
    GstPad* spad = NULL;
    GstElement *aparse = NULL;

#if 0
    aparse = gst_element_factory_make ("aacparse", "aparser");
    if (!aparse) {
      GST_ERROR("failed to create audio parse element");
      return;
    }
    gst_bin_add (srcbin->srcbin, aparse);

    if (!gst_element_link (aparse, srcbin->aqueue)) {
      GST_ERROR ("failed to link aparse & aqueue...\n");
      return;
    }
#endif

    g_print (" =========== >>>>>>>>>> Received AUDIO pad...\n");
    spad = gst_element_get_pad(GST_ELEMENT(client->srcbin->aqueue), "sink");
    if (gst_pad_link(pad, spad)  != GST_PAD_LINK_OK) {
      GST_ERROR ("Failed to link audio demuxer pad & audio parse pad...\n");
      return;
    }
    gst_object_unref(spad);
  } else {
    GST_ERROR ("not handling.....\n\n\n");
  }

  g_free (name);
}

static gboolean
gst_rtsp_client_create_audio_capture_bin (GstRTSPClient * client, GstRTSPClientSrcBin *srcbin)
{
  GstElement *audiosrc = NULL;
  GstElement *acaps = NULL;
  GstElement *aenc = NULL;
  GstElement *aencsrccaps = NULL;
  int channels = 0;
  int freq = 0;
  gchar *acodec = NULL;

  /* create audio src element */
  audiosrc = gst_element_factory_make ("pulsesrc", "audiosrc");
  if (!audiosrc) {
    GST_ERROR_OBJECT (client, "failed to create audiosrc element");
    goto create_error;
  }
  g_object_set (audiosrc, "device", "alsa_output.0.analog-stereo.monitor", NULL);
  //g_object_set (audiosrc, "provide-clock", 0, NULL);
  //g_object_set (audiosrc, "latency-time", 1000, NULL);
  //g_object_set (audiosrc, "actual-latency-time", 1000, NULL);

  g_object_set (audiosrc, "is-live", 1, NULL);
  g_object_set (audiosrc, "do-timestamp", 1, NULL);
  /* create audio caps element */
  acaps  = gst_element_factory_make ("capsfilter", "audiocaps");
  if (NULL == acaps) {
    GST_ERROR_OBJECT (client, "failed to create audio capsilfter element");
    goto create_error;
  }

  if(client->cChanels == WFD_CHANNEL_2)
    channels = 2;
  else if(client->cChanels == WFD_CHANNEL_4)
    channels = 4;
  else if(client->cChanels == WFD_CHANNEL_6)
    channels = 6;
  else if(client->cChanels == WFD_CHANNEL_8)
    channels = 8;
  else
    channels = 2;

  if(client->cFreq == WFD_FREQ_44100)
    freq = 44100;
  else if(client->cFreq == WFD_FREQ_48000)
    freq = 48000;
  else
    freq = 44100;

  if(client->caCodec == WFD_AUDIO_LPCM) {
    g_object_set (G_OBJECT (acaps), "caps",
        gst_caps_new_simple ("audio/x-lpcm",
        "width", G_TYPE_INT, client->cBitwidth,
        "rate", G_TYPE_INT, freq,
        "channels", G_TYPE_INT, channels, NULL), NULL);
  } else if((client->caCodec == WFD_AUDIO_AAC) || (client->caCodec == WFD_AUDIO_AC3)) {
    g_object_set (G_OBJECT (acaps), "caps",
        gst_caps_new_simple ("audio/x-raw-int",
            "width", G_TYPE_INT, client->cBitwidth,
            "endianess", G_TYPE_INT, 1234,
            "signed", G_TYPE_BOOLEAN, TRUE,
            "depth", G_TYPE_INT, 16,
            "rate", G_TYPE_INT, freq,
            "channels", G_TYPE_INT, channels, NULL), NULL);
  }

  if(client->caCodec == WFD_AUDIO_AAC)
    acodec = g_strdup("ffenc_aac");
  else if(client->caCodec == WFD_AUDIO_AC3)
    acodec = g_strdup("savsenc_ac3");
  else{
    GST_ERROR_OBJECT (client, "Yet to support other than H264 format");
    goto create_error;
  }

  aenc = gst_element_factory_make (acodec, "audioenc");
  if (NULL == aenc) {
    GST_ERROR_OBJECT (client, "failed to create audio encoder element");
    goto create_error;
  }
  aencsrccaps = gst_element_factory_make ("capsfilter", "audioencsrccaps");
  if (NULL == aencsrccaps) {
    GST_ERROR_OBJECT (client, "failed to create audio capsilfter element");
    goto create_error;
  }
  g_object_set (G_OBJECT (aencsrccaps), "caps",
        gst_caps_new_simple ("audio/mpeg",
        "stream-format", G_TYPE_STRING, "adts", NULL), NULL);

  srcbin->aqueue  = gst_element_factory_make ("queue", "audio-queue");
  if (!srcbin->aqueue) {
    GST_ERROR_OBJECT (client, "failed to create audio queue element");
    goto create_error;
  }

  gst_bin_add_many (srcbin->srcbin, audiosrc, acaps, aenc, aencsrccaps, srcbin->aqueue, NULL);

  if (!gst_element_link_many (audiosrc, acaps, aenc, aencsrccaps, srcbin->aqueue, NULL)) {
    GST_ERROR_OBJECT (client, "Failed to link audio src elements...");
    goto create_error;
  }

  return TRUE;

create_error:
  return FALSE;
}

static gboolean
gst_rtsp_client_create_xvcapture_bin (GstRTSPClient * client, GstRTSPClientSrcBin *srcbin)
{
  GstElement *videosrc = NULL;
  GstElement *vcaps = NULL;
  gchar *vcodec = NULL;

  videosrc = gst_element_factory_make ("xvimagesrc", "videosrc");
  if (NULL == videosrc) {
    GST_ERROR_OBJECT (client, "failed to create xvimagesrc element");
    goto create_error;
  }

  /* create video caps element */
  vcaps = gst_element_factory_make ("capsfilter", "videocaps");
  if (NULL == vcaps) {
    GST_ERROR_OBJECT (client, "failed to create video capsilfter element");
    goto create_error;
  }

  GST_INFO_OBJECT (client, "picked xvimagesrc as video source");
  g_object_set (G_OBJECT (vcaps), "caps",
      gst_caps_new_simple ("video/x-raw-yuv",
          "width", G_TYPE_INT, client->cMaxWidth,
          "height", G_TYPE_INT, client->cMaxHeight,
          "format", GST_TYPE_FOURCC, GST_MAKE_FOURCC('S','N','1','2'),
          "framerate", GST_TYPE_FRACTION, 30, 1, NULL), NULL);

  if(client->cvCodec == WFD_VIDEO_H264)
    vcodec = g_strdup("omx_h264enc");
  else {
    GST_ERROR_OBJECT (client, "Yet to support other than H264 format");
    goto create_error;
  }

  srcbin->venc  = gst_element_factory_make (vcodec, "videoenc");
  if (!srcbin->venc) {
    GST_ERROR_OBJECT (client, "failed to create video encoder element");
    goto create_error;
  }
  g_object_set (srcbin->venc, "bitrate", client->bitrate, NULL);
  g_object_set (srcbin->venc, "byte-stream", 1, NULL);
  g_object_set (srcbin->venc, "append-dci", 1, NULL);


  srcbin->vqueue  = gst_element_factory_make ("queue", "video-queue");
  if (!srcbin->vqueue) {
    GST_ERROR_OBJECT (client, "failed to create video queue element");
    goto create_error;
  }

  // TODO: check whether queue is required to have default values

  gst_bin_add_many (srcbin->srcbin, videosrc, vcaps, srcbin->venc, srcbin->vqueue, NULL);
  if (!gst_element_link_many (videosrc, vcaps, srcbin->venc, srcbin->vqueue,NULL)) {
    GST_ERROR_OBJECT (client, "Failed to link video src elements...");
    goto create_error;
  }

  return TRUE;

create_error:
  return FALSE;
}

static gboolean
gst_rtsp_client_create_filesrc_bin (GstRTSPClient * client, GstRTSPClientSrcBin *srcbin)
{
  GstElement *vdec = NULL;
  GstElement *videosrc = NULL;
  GstElement *srcdemux = NULL;

  videosrc = gst_element_factory_make ("filesrc", "videosrc");
  if (NULL == videosrc) {
    GST_ERROR_OBJECT (client, "failed to create filesrc element");
    goto create_error;
  }
  g_object_set (videosrc, "location", client->infile, NULL);

  GST_INFO_OBJECT (client, "picked filesrc as video source");

  // TODO: need to add support for remaing demuxers
  srcdemux = gst_element_factory_make ("qtdemux", "demuxer");
  if (!srcdemux) {
    GST_ERROR_OBJECT (client, "failed to create demuxer element");
    goto create_error;
  }
  g_signal_connect (srcdemux, "pad-added", G_CALLBACK(pad_added), client);

  srcbin->vparse = gst_element_factory_make ("legacyh264parse", "parser");
  if (!srcbin->vparse) {
    GST_ERROR_OBJECT(client, "failed to create video parse element");
    goto create_error;
  }
  g_object_set (srcbin->vparse, "config-interval", 1, NULL);
  g_object_set (srcbin->vparse, "output-format", 1, NULL);

  vdec = gst_element_factory_make ("omx_h264dec", "video-dec");
  if (!vdec) {
    GST_ERROR_OBJECT(client, "failed to create video decoder element");
    goto create_error;
  }

  srcbin->venc = gst_element_factory_make ("omx_h264enc", "video-enc");
  if (!srcbin->venc) {
    GST_ERROR_OBJECT(client, "failed to create video encoder element");
    goto create_error;
  }
  g_object_set (srcbin->venc, "bitrate", client->bitrate, NULL);
  g_object_set (srcbin->venc, "byte-stream", 1, NULL);
  g_object_set (srcbin->venc, "append-dci", 1, NULL);

  srcbin->vqueue  = gst_element_factory_make ("queue", "video-queue");
  if (!srcbin->vqueue) {
    GST_ERROR_OBJECT (client, "failed to create video queue element");
    goto create_error;
  }

  srcbin->aqueue  = gst_element_factory_make ("queue", "audio-queue");
  if (!srcbin->aqueue) {
    GST_ERROR_OBJECT (client, "failed to create audio queue element");
    goto create_error;
  }

  gst_bin_add_many (srcbin->srcbin, videosrc, srcdemux, srcbin->vparse, vdec,
    srcbin->venc, srcbin->vqueue, srcbin->aqueue, NULL);

  gst_element_link (videosrc, srcdemux);

  if (!gst_element_link_many (srcbin->vparse, vdec, srcbin->venc, srcbin->vqueue, NULL)) {
    GST_ERROR_OBJECT(client,"failed to link vparse &vqueue...\n");
    goto create_error;
  }

  return TRUE;

create_error:
  return FALSE;

}
/**
* gst_rtsp_client_create_srcbin:
* @client: client object
*
* Creates the server source bin
*
* Returns: void
*/
static void
gst_rtsp_client_create_srcbin (GstRTSPClient * client)
{
  GstRTSPClientSrcBin *srcbin = NULL;
  GstElement *mux = NULL;
  GstElement *payload = NULL;
  GstPad *srcpad = NULL;

  srcbin = g_slice_new0 (GstRTSPClientSrcBin);

  /* create source bin */
  srcbin->srcbin = gst_bin_new ("srcbin");
  if (!srcbin->srcbin) {
    GST_ERROR_OBJECT (client, "failed to create source bin...");
    goto create_error;
  }

  /* create video src element */
  switch (client->videosrc_type) {
    /* using xvimagesrc */
    case WFD_INI_VSRC_XVIMAGESRC: {
      if (!gst_rtsp_client_create_xvcapture_bin (client, srcbin)) {
        GST_ERROR_OBJECT (client, "failed to create xvcapture bin...");
        goto create_error;
      }
    }
    break;
    case WFD_INI_VSRC_FILESRC: {
      /* using filesrc */
      if (!gst_rtsp_client_create_filesrc_bin (client, srcbin)) {
        GST_ERROR_OBJECT (client, "failed to create xvcapture bin...");
        goto create_error;
      }
    }
    break;
    case WFD_INI_VSRC_CAMERASRC:
    case WFD_INI_VSRC_VIDEOTESTSRC:
    default:
      GST_ERROR_OBJECT (client, "unknow mode selected...");
      goto create_error;
  }

  mux = gst_element_factory_make ("mpegtsmux", "tsmux");
  if (!mux) {
    GST_ERROR_OBJECT (client, "failed to create muxer element");
    goto create_error;
  }
  if (client->hdcp_support) {
    g_object_set(mux, "HDCP-IP", client->wfdsink_ip, NULL);
    g_object_set(mux, "HDCP-port", client->hdcp_tcpport, NULL);
  }

  payload = gst_element_factory_make ("rtpmp2tpay", "pay0");
  if (!payload) {
    GST_ERROR_OBJECT (client, "failed to create payload element");
    goto create_error;
  }

  g_object_set (payload, "pt", 33, NULL);
  g_object_set (payload, "mtu", client->MTUsize, NULL);

  gst_bin_add_many (srcbin->srcbin, mux, payload, NULL);

  if (!gst_element_link_many (mux, payload, NULL)) {
    GST_ERROR_OBJECT (client, "Failed to link muxer & payload...");
    goto create_error;
  }

  /* request sink pad from muxer */
  srcbin->mux_vsinkpad = gst_element_get_request_pad (mux, "sink_%d");
  if (!srcbin->mux_vsinkpad) {
    GST_ERROR_OBJECT (client, "Failed to get sink pad from muxer...");
    goto create_error;
  }

  /* request srcpad from video queue */
  srcpad = gst_element_get_static_pad (srcbin->vqueue, "src");
  if (!srcpad) {
    GST_ERROR_OBJECT (client, "Failed to get srcpad from video queue...");
    goto create_error;
  }

  if (gst_pad_link (srcpad, srcbin->mux_vsinkpad) != GST_PAD_LINK_OK) {
    GST_ERROR_OBJECT (client, "Failed to link video queue src pad & muxer video sink pad...");
    goto create_error;
  }

  gst_object_unref (srcpad);
  srcpad = NULL;

  if (client->session_mode != WFD_INI_VIDEO_ONLY) {

    if (client->videosrc_type != WFD_INI_VSRC_FILESRC) {
      /* create audio source elements & add to pipeline */
      if (!gst_rtsp_client_create_audio_capture_bin (client, srcbin))
        goto create_error;
    }

    /* request sink pad from muxer */
    srcbin->mux_asinkpad = gst_element_get_request_pad (mux, "sink_%d");
    if (!srcbin->mux_asinkpad) {
      GST_ERROR_OBJECT (client, "Failed to get sinkpad from muxer...");
      goto create_error;
    }

    /* request srcpad from audio queue */
    srcpad = gst_element_get_static_pad (srcbin->aqueue, "src");
    if (!srcpad) {
      GST_ERROR_OBJECT (client, "Failed to get srcpad from audio queue...");
      goto create_error;
    }

    /* link audio queue's srcpad & muxer sink pad */
    if (gst_pad_link (srcpad, srcbin->mux_asinkpad) != GST_PAD_LINK_OK) {
      GST_ERROR_OBJECT (client, "Failed to link audio queue src pad & muxer audio sink pad...");
      goto create_error;
    }

#ifdef WFD_PAD_PROBE
    GstPad *pad_probe = NULL;
    pad_probe = gst_element_get_static_pad (mux, "src");
    if(NULL == pad_probe)
    {
      GST_INFO("pad for probe not created");
    } else
    GST_INFO("pad for probe SUCCESSFUL");
    gst_pad_add_data_probe (pad_probe, G_CALLBACK (gst_dump_data), client);
#endif

    gst_object_unref (srcpad);
    srcpad = NULL;
  }

  client->srcbin = srcbin;

  GST_DEBUG_OBJECT (client, "successfully created source bin...");

  return;

create_error:
  g_slice_free (GstRTSPClientSrcBin, srcbin);

  // TODO: Need to clean everything
  return;
}

#ifdef WFD_PAD_PROBE
static gboolean gst_dump_data (GstPad * pad, GstMiniObject * obj, gpointer u_data)
{
  gint8 *data;
  gint size;
  GstElement *bin = (GstElement *) u_data;
  if (GST_IS_BUFFER (obj)) {
    GstBuffer *buffer = GST_BUFFER_CAST (obj);
    GST_LOG ("got buffer %p with size %d", buffer, GST_BUFFER_SIZE (buffer));
    data = GST_BUFFER_DATA(buffer);
    size = GST_BUFFER_SIZE(buffer);
    f = fopen("probe.ts", "a");
    fwrite(data, size, 1, f);
    fclose(f);
  }
  return TRUE;
}
#endif

static void
keep_alive_response_check (gpointer *userdata)
{
  GstRTSPClient *client = (GstRTSPClient *)userdata;
  if (!client) {
    return;
  }
  if (client->keep_alive_flag) {
    return;
  }
  else {
    GST_INFO ("%p: source error notification", client);
    g_signal_emit (client, gst_rtsp_client_signals[SIGNAL_ERROR], 0, NULL);
  }
}

/*CHecking whether source has got response of any request.
 * If yes, keep alive message is sent otherwise error message
 * will be displayed.*/
gboolean
keep_alive_condition(gpointer *userdata)
{
  GstRTSPClient *client;
  gboolean res;
  client = (GstRTSPClient *)userdata;
  if (!client) {
    return FALSE;
  }
  if(client->keep_alive_lock)
    g_mutex_lock(client->keep_alive_lock);
  else
    return FALSE;
  if(!client->keep_alive_flag) {
    g_timeout_add(5000, keep_alive_response_check, client);
  }
  else {
    GST_DEBUG_OBJECT (client, "have received last keep alive message response");
  }
  GST_DEBUG("sending keep alive message");
  res = gst_rtsp_client_sending_m16_message(client);
  if(res) {
    client->keep_alive_flag = FALSE;
  }
  else {
    GST_ERROR_OBJECT (client, "Failed to send Keep Alive Message");
    g_mutex_unlock(client->keep_alive_lock);
    return FALSE;
  }
  g_mutex_unlock(client->keep_alive_lock);
  return TRUE;
}
/*Sending keep_alive (M16) message.
  Without calling prepare_request function.*/

gboolean
gst_rtsp_client_sending_m16_message (GstRTSPClient * client)
{
  GstRTSPResult res = GST_RTSP_OK;
  GstRTSPUrl *url = NULL;
  GstRTSPMessage request = { 0 };
  gchar *url_str = NULL;
  GstRTSPSession *session = NULL;

  url = gst_rtsp_connection_get_url (client->connection);
  if (url == NULL) {
    GST_ERROR_OBJECT (client, "Failed to get connection URL");
    return FALSE;
  }

  url_str = gst_rtsp_url_get_request_uri (url);
  if (url_str == NULL) {
    GST_ERROR_OBJECT (client, "Failed to get connection URL");
    return FALSE;
  }

  res = gst_rtsp_message_init_request (&request, GST_RTSP_GET_PARAMETER, url_str);
  if (res < 0) {
    GST_ERROR ("init request failed");
    return FALSE;
  }

  if (client->sessionid) {
    session = gst_rtsp_session_pool_find (client->session_pool, client->sessionid);
    GST_INFO_OBJECT (client, "session = %p & sessionid = %s", session, session->sessionid);
  }
  send_request (client, session, &request);
  return TRUE;
}

